/* Project mfs --- Migration file system
 * mfs_com.c -- User/kernel communication implementation
 * Creation: 17/10/1999
 * Update:   24/02/2000
 * V1.0 PV003
 * AUTHOR
 *    Marc SCHAEFER <schaefer@alphanet.ch>
 * LICENSE
 *    GPL
 * DESCRIPTION
 *    This implements the communication methods with the user process
 *    mfsd.
 * NOTES
 *    - Current implementation assumes the following:
 *         - wake_up() does NOT schedule(): atomicity is guaranteed till
 *           the process exits kernel mode, calls sleep_on(), or accesses
 *           user level data.
 *         - wake_up() wakes up all waiting processes of that queue (this is
 *           however not required by our implementation, we presumably would
 *           be satisfied by wake_up_one_only()). This concerns only the
 *           no_slot_free queue (all others have max one process at this time).
 *      All the above is expected to be true also on SMP systems (!?). And
 *      there is no risk of bad caching, etc ??
 *    - Only one guy is allowed to open the fs for mfsd ioctls(). When
 *      the mfsd one closes mfs_notify_dies() must be called. If none
 *      are opened, no requests should be queued.
 * BUGS
 *    - SMP might need changes.
 * TODO
 *    - Function mfs_shutdown() called from fs when unmounting, which sets
 *      shutdown_requested BOOL, and awaits mfsd closing the file on a
 *      special queue.
 * BASED-ON
 *    Original work.
 * MODIFICATION-HISTORY
 *    17/10/99  schaefer  Created this file.
 *    11/02/00  schaefer  Cleanup + replaced debugging by FIST debugging.
 *    24/02/00  schaefer  Added poll() support.
 *    26/02/00  schaefer  No longer relying on clearing the queue.
 * $Id: mfs_com.c,v 1.39 2000/02/27 18:12:15 schaefer Exp $
 */

#include <linux/kernel.h>
#include <linux/wait.h>
#include <linux/stat.h>
#include <linux/malloc.h>
#include <linux/fs.h>
#include <linux/poll.h>
#include <asm/uaccess.h>
#include "mfs_definitions.h"
#include "mfs_interface.h"
#include "mfs_com_interface.h"
#include "fist.h" /* for fist_dprint() */

/* Private type definitions */

#define MFS_COM_MAX_PENDING_REQUESTS 8
#define MFS_COM_RESERVED_ENTRIES 1 /* for index queries */
typedef int mfs_com_request_count_t;

typedef struct {
   struct wait_queue *wait;
   mfs_request_t request;
   mfs_answer_t answer;
   BOOL mfsd_running; /* request running in mfsd */
} mfs_com_request_t;

typedef struct {
   mfs_com_request_t request[MFS_COM_MAX_PENDING_REQUESTS];
   mfs_com_request_count_t count; /* @@@ is that really used ? @@@ */
   struct wait_queue *no_free_request;

   struct wait_queue *mfsd;

   mfs_dentry_t *mfs_special_dentry;
   struct task_struct *mfs_task_id;

   BOOL shutdown_requested;
   BOOL mfsd_died;
} mfs_com_private_data_t;

/* Private function definitions */

/* NAME
 *    mfs_get_slot
 * DESCRIPTION
 *    Returns either an empty slot for kernel usage, or a full slot for
 *    mfsd usage (with mfsd_running FALSE). This also accounts for
 *    reserved slots if !mfsd.
 *    If none of that type are available, returns NULL.
 * RESULT
 *    mfs_com_request_t *, NULL if no request of that type.
 * NOTES
 *    - request_number can be NULL and won't be set in that case.
 *    - Will not modify the request queue.
 * BUGS
 * TODO
 */
static mfs_com_request_t *mfs_get_slot(mfs_com_private_data_t *data,
                                       mfs_com_request_count_t *request_number,
                                       BOOL mfsd,
                                       BOOL reserved_ok);

/* NAME
 *    mfs_return 
 * DESCRIPTION
 *    Returns the specified active request to the user with the specified
 *    error status.
 * RESULT
 *    NONE
 * NOTES
 *    - status is always expected to be MFS_STATUS_MFSD_UNAVAILABLE.
 * BUGS
 * TODO
 */
static void mfs_return(mfs_com_request_t *request, mfs_status_t status);

/* NAME
 *    dentry_get_path
 * DESCRIPTION
 *    Writes the complete NUL terminated path, including /,
 *    to the given buffer, in the usual order. The function
 *    doesn't sleep. The buffer is a kernel-space buffer,
 *    which can be accessed directly.
 * RESULT
 *    0        operation succeeded.
 *    -ENOMEM  buffer was too small, buffer can't be considered.
 *    -EIO     integrity failure when evaluating path, buffer can't be
 *             considered.
 * NOTES
 *    - We do not lock the dentries since we assume we will not
 *      sleep at all. This is maybe wrong in SMP.
 *    - We remove all the leading /.
 *    - The reason we don't use d_path is multiple:
 *         - d_path is badly documented
 *         - the conditions where d_path adds a " (deleted)" are not
 *           really clear. At this when THIS function was designed,
 *           we were working on unexistant underlying inodes (since then,
 *           we simplified the design).
 *         - the memcpy() would be needed anyway, or the interface to mfsd
 *           changed with a specified offset.
 *    - supports a NULL dentry (will return a NUL path).
 * BUGS
 * TODO
 */
static int dentry_get_path(mfs_dentry_t *dentry,
                           char *buffer,
                           int buffer_length);

/* NAME
 *    mfs_set_current_ia_attr
 * DESCRIPTION
 *    Sets the current attributes flags for inode in ATTR_ format.
 * RESULT
 *    NONE
 * NOTES
 * BUGS
 *    - Implementation is presumably wrong, since inode->i_flags
 *      does not correspond 1-1 to ATTR_FLAGs.
 * TODO
 */
static void mfs_set_current_ia_attr(const mfs_dentry_t *dentry,
                                    struct iattr *to_set);

/* NAME
 *    mfsd_linked
 * DESCRIPTION
 *    Returns TRUE if mfsd is currently linked on the current process
 *    and current file. This means that MFS_IOCTL_NOTIFY_LIVES was called
 *    before.
 * RESULT
 *    TRUE if mfsd correctly linked.
 * NOTES
 * BUGS
 * TODO
 */
static BOOL mfsd_linked(mfs_dentry_t *dentry, mfs_com_private_data_t *data);

/* Private variables */

/* Function implementation */

void *mfs_initialize(void) {
   mfs_com_private_data_t *data;

   fist_dprint(6, "mfs_initialize().\n");

   data = kmalloc(sizeof(*data), GFP_KERNEL);
   if (data) {
      mfs_com_request_count_t i;

      data->count = 0;
      data->no_free_request = NULL;
      data->mfsd = NULL;
      data->mfs_special_dentry = NULL;
      data->mfs_task_id = NULL;
      data->shutdown_requested = FALSE;
      data->mfsd_died = FALSE; /* @@ should be set to TRUE, set to false when
                                * mfsd opens the file and does the first ioctl,
                                * and then to false through call of notify
                                */

      for (i = 0; i < MFS_COM_MAX_PENDING_REQUESTS; i++) {
         data->request[i].wait = NULL;
         data->request[i].mfsd_running = FALSE;
      }

      return (void *) data;
   }
   else {
      return NULL;
   }
}

void mfs_deinitialize(void *private) {
   mfs_com_request_count_t i;
   mfs_com_private_data_t *data;
   mfs_com_request_count_t count = 0;

   fist_dprint(6, "mfs_deinitialize().\n");

   data = (mfs_com_private_data_t *) private;

   for (i = 0; i < MFS_COM_MAX_PENDING_REQUESTS; i++) {
      if (waitqueue_active(&data->request[i].wait)) {
         count++;
         mfs_return(&data->request[i], MFS_STATUS_MFSD_UNAVAILABLE);
      }
   }

   /* For all in the no slot wait list, awake them. They will notice
    * the shutdown.
    * BUGS
    *    - we should really wait so they can see the shutdown flag ...
    */
   wake_up(&data->no_free_request);

   if (count) {
      /* Can't happen since fs can't be closed with one file open
       * BUGS
       *    - well, blocked before open because migrated??!
       */
      printk("mfs: mfs_com: still %d pending requests; cancelled.\n", count);
   }

   if (waitqueue_active(&data->mfsd)) {
      /* Can't happen since fs can't be closed with one file open */
      printk("mfs: mfs_com: mfsd still running ?!.\n");
      wake_up(&data->mfsd); /* mark mfsd for scheduling */
   }

   kfree(data);
}

void mfs_notify_dies(mfs_dentry_t *dentry, void *private) {
   mfs_com_request_count_t i;
   mfs_com_private_data_t *data;
 
   data = (mfs_com_private_data_t *) private;

   if (mfsd_linked(dentry, data)) {
      data->mfs_special_dentry = NULL;
      data->mfs_task_id = NULL;

      fist_dprint(6, "mfs_notify_dies().\n");

      for (i = 0; i < MFS_COM_MAX_PENDING_REQUESTS; i++) {
	 if (waitqueue_active(&data->request[i].wait)) {
	    mfs_return(&data->request[i], MFS_STATUS_MFSD_UNAVAILABLE);
	 }
      }

      /* Wake up all processes waiting for free slot, setting the
       * mfsd_died.
       */
      data->mfsd_died = TRUE;
      wake_up(&data->no_free_request);  /* note that wake_up doesn't clear the
					 * queue. Return from sleep_on() of
					 * other process does.
					 */
   }
}

int mfs_notify_lives(mfs_dentry_t *dentry, void *private, void *user_space) {
   mfs_com_private_data_t *data;
   int result = -EIO;

   data = (mfs_com_private_data_t *) private;

   if (dentry && !data->mfs_special_dentry && !data->mfs_task_id) {
      char *buffer;

      buffer = kmalloc(MAX_PATHNAME_LENGTH, GFP_KERNEL); /* can sleep */
      if (buffer) {
	 int res;

	 res = dentry_get_path(dentry,
			       buffer,
			       MAX_PATHNAME_LENGTH);
	 if (res) {
	    printk("mfs: mfs_notify_lives: %s [%d].\n",
		   (res == -ENOMEM) ? "path too long"
				    : "internal error",
		   res);
	 }
         else {
	    if (!strcmp(buffer, MFS_SPECIAL_MFSD_FILE)) {
               int pversion;

               result = get_user(pversion, (int *) user_space);
               if (result == 0) {
		  if (pversion == MFS_PROTOCOL_VERSION) {
		     mfs_info_t mfs_info;

		     mfs_info.protocol_version = MFS_PROTOCOL_VERSION;
		     mfs_info.queue_depth = MFS_COM_MAX_PENDING_REQUESTS;
		     mfs_info.reserved_entries = MFS_COM_RESERVED_ENTRIES;

		     if (copy_to_user(user_space,
                                      &mfs_info,
                                      sizeof(mfs_info))) {
			result = -EFAULT;
		     }
		     else {
			data->mfsd_died = FALSE;

			fist_dprint(6, "mfs_notify_lives: LIVES!.\n");

			data->mfs_special_dentry = dentry;
			data->mfs_task_id = current;

                        /* result 0 already */
		     }
		  }
		  else {
		     result = -ENODEV;
		  }
               }
	    }
         }

         kfree(buffer);
      }
      else {
         printk("mfs_notify_lives: no memory.\n");
         result = -ENOMEM;
      }
   }

   return result;
}

BOOL mfs_is_special_poll(file_t *file,
                         poll_table *wait,
                         void *private,
                         int *mask) {
   mfs_com_private_data_t *data;
   BOOL result = FALSE;
   mfs_dentry_t *dentry = file->f_dentry;

   data = (mfs_com_private_data_t *) private;

   if (dentry
       && mfsd_linked(dentry, data)) {
      *mask = 0;

      /* NOTES
       *    - We rely on the assumption that mfs_get_slot() doesn't
       *      modify the request queue.
       */
      if (mfs_get_slot(data, NULL, TRUE, TRUE /* not really important */)) {
         /* Exceptionnal data available (ioctl()) */
         *mask |= POLLPRI;
      }
      else {
         /* Add to wait queue */
         poll_wait(file, &data->mfsd, wait);
      }
      result = TRUE;
   }

   return result;
}

mfs_status_t mfs_do_request(void *private,
                            mfs_command_t command,
                            mfs_dentry_t *dentry,
                            mfs_index_object_size_t *size,
                            void *special) {
   mfs_com_private_data_t *data;
   mfs_status_t result = MFS_STATUS_OK;
   BOOL done = FALSE;

   fist_dprint(6, "mfs_do_request().\n");

   data = (mfs_com_private_data_t *) private;

   do {
      mfs_com_request_count_t i;
      mfs_com_request_t *rq;

      if (current == data->mfs_task_id) {
	 printk("mfs_do_request: DEADLOCK AVOIDED!\n");
	 result = MFS_STATUS_MFSD_UNAVAILABLE;
         break; /* out of while */
      }
      
      if (data->mfsd_died) {
         /* Note that if mfsd died, notify was called, and all running
          * requested were cancelled, and all processes waiting for a
          * slot were also awaken.
          */
         result = MFS_STATUS_MFSD_UNAVAILABLE;
         break; /* out of while */
      }
      
      rq = mfs_get_slot(data,
                        &i,
                        FALSE /* free */,
                        (command == MFS_COMMAND_GET_INDEX_OBJECT));
      if (rq) {
         if (command == MFS_COMMAND_CREATE_MIGRATED_OBJECT) {
            strncpy(rq->request.path_name,
                    ((mfs_import_t *) special)->file_name,
                    sizeof(rq->request.path_name));
            rq->request.import.locator = ((mfs_import_t *) special)->locator;
            rq->request.import.size = ((mfs_import_t *) special)->size;
            rq->request.import.uid = ((mfs_import_t *) special)->uid;
            rq->request.import.gid = ((mfs_import_t *) special)->gid;
            rq->request.import.mtime = ((mfs_import_t *) special)->mtime;
            rq->request.import.mode = ((mfs_import_t *) special)->mode;
            strncpy(rq->request.import.migrated_on,
                    ((mfs_import_t *) special)->migrated_on,
                    sizeof(rq->request.import.migrated_on));
         }
         else if ((command == MFS_COMMAND_MIGRATE_OBJECT)
                  || ((command == MFS_COMMAND_UNMIGRATE_OBJECT)
                      && special)) {
            strncpy(rq->request.path_name,
                    ((mfs_migrate_unmigrate_t *) special)->file_name,
                    sizeof(rq->request.path_name));
         }
         else if (command == MFS_COMMAND_SHUTDOWN) {
            rq->request.path_name[0] = '\0';
         }
         else {
            int res;

	    res = dentry_get_path(dentry,
				  rq->request.path_name,
				  sizeof(rq->request.path_name));
	    if (res) {
	       printk("mfs: mfs_com: %s [%d].\n",
		      (res == -ENOMEM) ? "path too long"
				       : "internal error",
		      res);
	       result = MFS_STATUS_OBJECT_NOT_FOUND;
	       goto out_abort;
	    }
         }

         fist_dprint(6, "mfs_com: complete name: %s.\n",
                rq->request.path_name);

         /* This is not exactly fair scheduling, especially for
          * the ones in the no_free_request queue. Think again.
          */
         rq->request.command = command;
         rq->request.private = (void *) i; /* tracking from mfsd */
               
         rq->mfsd_running = FALSE;
         if (waitqueue_active(&data->mfsd)) {
            wake_up(&data->mfsd); /* mark mfsd for scheduling */
         }

         /* else it's there, file open, but not in the ioctl yet */
         sleep_on(&rq->wait);

         done = TRUE;
      
         result = rq->answer.status;
      
         /* Here fill the inode, etc, WITHOUT TOUCHING USERSPACE, 
          * or sleeping, since we already cleared the .wait and will be
          * reused.
          */
      
         if (size) {
            *size = rq->answer.size;
         }
      
         wake_up(&data->no_free_request);
      }
      else {
         sleep_on(&data->no_free_request);
         /* Instead of having those two different queues, we could have
          * allocated the request on the stack. However, this only works
          * when data to be passed is small. As we pass big pathnames and
          * stat_buff around, we must limit the stack usage.
          */
      }
   } while (!done);

out_abort:
   return result;
}

int mfs_transmit_request(void *private,
                         void *user_space,
                         mfs_dentry_t *dentry) {
   int err = 0;
   mfs_com_private_data_t *data = (mfs_com_private_data_t *) private;

   if (mfsd_linked(dentry, data)) {
      mfs_com_request_t *request;
      BOOL no_protocol = FALSE;

      do {
	 request = mfs_get_slot(data, NULL, TRUE, TRUE /* not important */);
	 if (request) {
	    request->mfsd_running = TRUE;
	    if (copy_to_user(user_space,
			     &request->request,
			     sizeof(request->request))) {
	       err = -EFAULT;
	    }
	 }
	 else {
	    sleep_on(&data->mfsd);
	 }

	 no_protocol = data->shutdown_requested;
      } while ((request == NULL) && !no_protocol);

      if (err == 0) {
	 if (no_protocol) {
	    err = -EIO;
	 }
      }
   }
   else {
      err = -EIO;
      printk("mfs_transmit_request: missing MFS_IOCTL_NOTIFY_LIVES.\n");
   }

   return err;
}

int mfs_receive_answer(void *private,
                       void *user_space,
                       mfs_dentry_t *dentry) {
   int err = 0;
   mfs_com_private_data_t *data = (mfs_com_private_data_t *) private;

   if (mfsd_linked(dentry, data)) {
      mfs_answer_t answer;
      mfs_com_request_count_t request_number;

      if (copy_from_user(&answer, user_space, sizeof(answer))) {
	  err = -EFAULT;
      }
      else {
	 request_number = (mfs_com_request_count_t) answer.private;

	 if ((request_number < MFS_COM_MAX_PENDING_REQUESTS)
	     && (request_number >= 0)) {
	    memcpy(&data->request[request_number].answer,
		   &answer,
		   sizeof(data->request[request_number].answer));
	    if (!data->request[request_number].mfsd_running) {
	       printk("mfs: mfs_receive_answer but not running ?!.\n");
	    }
	    if (waitqueue_active(&data->request[request_number].wait)) {
	       wake_up(&data->request[request_number].wait);
	    }
	    else {
	       printk("mfs: mfs_receive_answer: request not active ?!.\n");
	    }
	 }
	 else {
	    err = -ENOENT;
	    printk("mfs: mfs_receive_answer: request not found ?!.\n");
	 }
      }
   }
   else {
      err = -EIO;      
      printk("mfs_receive_answer: missing MFS_IOCTL_NOTIFY_LIVES.\n");
   }

   return err;
}

BOOL mfs_dentry_in_file(mfs_dentry_t *dentry) {
   struct file *file = inuse_filps;

   /* BUGS
    *    - Is that safe and right and the most efficient way ?
    *    - Needs inuse_filps to be exported from kernel.
    */

   while (file) {
      if (file->f_dentry == dentry) {
         return TRUE;
      }
      file = file->f_next;   
   }

   return FALSE;
}

void mfs_dump_all_fs_files(struct super_block *sb) {
   char *buffer;
   
   buffer = kmalloc(MAX_PATHNAME_LENGTH, GFP_KERNEL); /* can sleep */
   if (buffer) {
      struct file *file = inuse_filps;
      /* BUGS
       *    - Is that safe and right and the most efficient way ?
       *    - I think it's especially wrong because of the
       *      printk()s.
       *    - Needs inuse_filps to be exported from kernel.
       */

      printk("mfs_dump_all_files START\n");
      while (file) {
	 if (file->f_dentry->d_sb == sb) {
	    int result;

	    result = dentry_get_path(file->f_dentry,
                                     buffer,
                                     MAX_PATHNAME_LENGTH);
	    if (result == 0) {
	       printk("%s %d %d\n",
                      buffer,
                      file->f_dentry->d_count,
                      file->f_dentry->d_inode->i_count);
	    }
	    else {
	       printk("mfs_dump_all_fs_files: error %d on dentry 0x%lx.\n",
		      result,
		      (unsigned long int) file->f_dentry);
	    }
	 }
	 file = file->f_next;   
      }
      printk("mfs_dump_all_files END.\n");

      kfree(buffer);
   }
   else {
      printk("mfs_dump_all_fs_files: no memory.\n");
   }
}

int mfs_clear_migrated_flag(mfs_dentry_t *dentry) {
   struct iattr newattrs;

   mfs_set_current_ia_attr(dentry, &newattrs);
   newattrs.ia_attr_flags &= ~(ATTR_FLAG_NOATIME);
   newattrs.ia_valid = ATTR_ATTR_FLAG;

   return notify_change(dentry, &newattrs);
}

int mfs_set_migrated_flag(mfs_dentry_t *dentry) {
   struct iattr newattrs;

   mfs_set_current_ia_attr(dentry, &newattrs);
   newattrs.ia_attr_flags |= ATTR_FLAG_NOATIME;
   newattrs.ia_valid = ATTR_ATTR_FLAG;

   return notify_change(dentry, &newattrs);
}

/* Private function implementation */

static mfs_com_request_t *mfs_get_slot(mfs_com_private_data_t *data,
                                       mfs_com_request_count_t *request_number,
                                       BOOL mfsd,
                                       BOOL reserved_ok) {
   mfs_com_request_count_t i;
   BOOL found = FALSE;

   /* This is not round-robin scheduling. */

   for (i = (MFS_COM_MAX_PENDING_REQUESTS - 1); i >= 0; i--) { /* backwards */
      if ((waitqueue_active(&data->request[i].wait)
          && !data->request[i].mfsd_running
          && mfsd)
         ||
         (!waitqueue_active(&data->request[i].wait)
          && !mfsd)) {
         found = TRUE;
         break;
      }
   }

   if (found && (mfsd || reserved_ok || (i >= MFS_COM_RESERVED_ENTRIES))) {
      if (request_number) {
	 *request_number = i;
      }

      return &data->request[i];
   }
   else {
      return NULL;
   }
}

static void mfs_return(mfs_com_request_t *request, mfs_status_t status) {
   if (status != MFS_STATUS_MFSD_UNAVAILABLE) {
      printk("mfs: mfs_return: incorrect status 0x%x.\n", status);
   }

   if (waitqueue_active(&request->wait)) {
      request->answer.status = status;
      wake_up(&request->wait);
   }
   else {
      printk("mfs: mfs_return: request not active ?!.\n");
   }
}

static int dentry_get_path(mfs_dentry_t *dentry,
                           char *buffer,
                           int buffer_length) {
   char *after_buffer_address = buffer + buffer_length; /* is OUTSIDE ! */
   char *last_address = after_buffer_address;

   while (dentry) {
      int element_size = dentry->d_name.len + 1;

      /* The +1 either accounts for the preceeding / or the last NUL */

      if (element_size == 1) {
         return -EIO;
      }
      else if (element_size > buffer_length) {
         return -ENOMEM;
      }
      else {
         after_buffer_address -= element_size;
         /* Thus, we now reserved from current..last_current */
         buffer_length -= element_size;
         memcpy(after_buffer_address + 1,
                dentry->d_name.name,
                dentry->d_name.len);
         *after_buffer_address = '/';
      }

      if (dentry == dentry->d_parent) {
         dentry = NULL;
      }
      else {
         dentry = dentry->d_parent;
      }
   }

   /* Now shift the buffer, ignoring all the first / */

   while ((after_buffer_address < last_address)
          && (*after_buffer_address == '/')) {
      after_buffer_address++;
   }

   /* We could have used memcpy if we were sure it was supporting
    * overlapped copies.
    */
   while (after_buffer_address < last_address) {
      *buffer = *after_buffer_address;
      buffer++;
      after_buffer_address++;
   }

   *buffer = '\0'; /* and finish the buffer. There is room because of the
                    * initial removed /.
                    */

   return 0;   
}

static void mfs_set_current_ia_attr(const mfs_dentry_t *dentry,
                                    struct iattr *to_set) {
   /* BUGS
    *    - see spec.
    */
   to_set->ia_attr_flags = dentry->d_inode->i_flags;
}

static BOOL mfsd_linked(mfs_dentry_t *dentry, mfs_com_private_data_t *data) {
   return (data->mfs_task_id == current)
          && (data->mfs_special_dentry == dentry);
}
