/* Project mfs --- Migration file system
 * mfs_com_interface.h -- kernel support routines for communication definitions
 * Creation: 19/10/1999
 * Update:   27/02/2000
 * V1.0 PV004
 * AUTHOR
 *    Marc SCHAEFER <schaefer@alphanet.ch>
 * LICENSE
 *    GPL
 * DESCRIPTION
 *    This include file defines types and functions exported for use by
 *    the mfs kernel driver when it wants to communicate with the mfsd
 *    process.
 * NOTES
 * BUGS
 * TODO
 * BASED-ON
 *    Original work.
 * MODIFICATION-HISTORY
 *    19/10/99  schaefer  Created this file.
 *    11/02/00  schaefer  Removed unneeded debugging.
 *    24/02/00  schaefer  Added poll() support.
 *    27/02/00  schaefer  notify_lives() sets mfs_info_t.
 * $Id: mfs_com_interface.h,v 1.23 2000/02/27 18:12:15 schaefer Exp $
 */

#ifndef _MFS_COM_INTERFACE_H
#define _MFS_COM_INTERFACE_H

#include "mfs_user_ioctls.h"

/* Type definitions */

typedef struct dentry mfs_dentry_t;

/* Macros */

#define MFS_CLEAR_MIGRATED_FLAG(dentry) mfs_clear_migrated_flag(dentry);

#define MFS_IS_MIGRATED_FLAG(inode) ((inode)->i_flags & MS_NOATIME)

/* Does not change the dentry ctime */
#define MFS_SET_MIGRATED_FLAG(dentry) mfs_set_migrated_flag(dentry);

/* Function definitions */

/* NAME
 *    mfs_initialize
 * DESCRIPTION
 *    Initializes the private mfs_com data structures.
 * RESULT
 *    void *, to be stored somewhere in the filesystem private
 *    data and to be passed for every following mfs_* call.
 *    Can be NULL if the initialization failed.
 * NOTES
 *    - There must be one call to mfs_deinitialize() for every call to
 *      mfs_initialize() where the result was non NULL. Calls should not nest.
 * BUGS
 * TODO
 */
void *mfs_initialize(void);

/* NAME
 *    mfs_deinitialize
 * DESCRIPTION
 *    Deinitializes (and possibly frees) the private mfs_com data structures.
 * RESULT
 *    NONE
 * NOTES
 * BUGS
 * TODO
 */
void mfs_deinitialize(void *private);

/* NAME
 *    mfs_do_request
 * DESCRIPTION
 *    Executes a request to the mfsd process, waiting for the answer by
 *    suspending the current task.
 * RESULT
 *    mfs_status_t
 * NOTES
 *    - status will be set depending on what mfsd says. If no mfsd
 *      is available, OR IF mfsd IS KILLED during the operation, a special
 *      status of MFS_STATUS_MFSD_UNAVAILABLE is returned.
 *    - This function will be called with a user context, in kernel mode,
 *      with scheduling disabled.
 *    - If the string is too long to fit in mfsd intercommunication,
 *      we return MFS_STATUS_OBJECT_NOT_FOUND.
 *    - size can be set to NULL and no data will be returned (useful
 *      e.g. when no data is to be returned).
 *    - dentry can also be NULL, except for MFS_COMMAND_GET_INDEX_OBJECT
 *      and MFS_COMMAND_UNMIGRATE_OBJECT.
 *    - dentry is also used for saving the mfsd's special file dentry.
 *    - special can be NULL if command is not
 *      MFS_COMMAND_CREATE_MIGRATED_OBJECT, MFS_COMMAND_UNMIGRATE_OBJECT
 *      or MFS_COMMAND_UNMIGRATE_OBJECT.
 * BUGS
 *    - A little messy. Merge the different type of calls (normal or
 *      MFS_COMMAND_CREATE_MIGRATED_OBJECT).
 * TODO
 */
mfs_status_t mfs_do_request(void *private,
                            mfs_command_t command,
                            mfs_dentry_t *dentry,
                            mfs_index_object_size_t *size,
                            void *special);

/* NAME
 *    mfs_transmit_request
 * DESCRIPTION
 *    Fetches a request from the queue, and copies it to the user-space
 *    pointer. Sleeps if no requests are available.
 * RESULT
 *    0 if ok, -EFAULT if user space access error, -EIO if filesystem
 *    detached.
 * NOTES
 * BUGS
 * TODO
 */
int mfs_transmit_request(void *private,
                         void *user_space,
                         mfs_dentry_t *dentry);

/* NAME
 *    mfs_receive_answer
 * DESCRIPTION
 *    Matches the correct answer to the queue, and copies it to the queue.
 * RESULT
 *    0 if ok, -EFAULT if user space access error, -EIO if filesystem
 *    detached, -ENOENT if no matching queue entry.
 * NOTES
 * BUGS
 * TODO
 */
int mfs_receive_answer(void *private,
                       void *user_space,
                       mfs_dentry_t *dentry);

/* NAME
 *    mfs_notify_dies
 * DESCRIPTION
 *    Informs the mfsd process has closed the ioctl() file descriptor and
 *    thus any running command should be returned with error status.
 * RESULT
 *    NONE
 * NOTES
 *    - This function will be called with a user context, in kernel mode,
 *      with scheduling disabled.
 *    - Infact the function decides based on the dentry.
 * BUGS
 * TODO
 */
void mfs_notify_dies(mfs_dentry_t *dentry, void *private);

/* NAME
 *    mfs_is_special_poll
 * DESCRIPTION
 *    If this is the special file and the current process is mfsd,
 *    we add support for the poll() method.
 * RESULT
 *    TRUE if this was the case
 * NOTES
 *    - If returns FALSE, no variable was touched. You should
 *      thus forward the request to the underlying fs.
 * BUGS
 * TODO
 */
BOOL mfs_is_special_poll(struct file *file,
                         poll_table *wait,
                         void *private,
                         int *mask);

/* NAME
 *    mfs_notify_lives
 * DESCRIPTION
 *    Informs that the current task is a a mfsd process.
 * RESULT
 *    0 if ok, -EFAULT if user space access error, -EIO if mfsd already
 *    attached, -ENODEV if wrong protocol version, -ENOMEM if can't attach.
 * NOTES
 *    - This function will be called with a user context, in kernel mode,
 *      with scheduling disabled.
 *    - The function may cause scheduling to happen.
 *    - Infact the function decides based on the file name and of the
 *      fact of if there already is a mfsd.
 *    - Parameter coming from user space is mfs_info.protocol_version,
 *      the whole structure is set by kernel.
 * BUGS
 * TODO
 */
int mfs_notify_lives(mfs_dentry_t *dentry, void *private, void *user_space);

/* NAME
 *    mfs_dentry_in_file
 * DESCRIPTION
 *    Returns TRUE if a struct file * pointing to the dentry exists.
 * RESULT
 *    BOOL
 * NOTES
 * BUGS
 * TODO
 */
BOOL mfs_dentry_in_file(mfs_dentry_t *dentry);

/* NAME
 *    mfs_dump_all_fs_files
 * DESCRIPTION
 *    Dumps to the kernel debug buffer all the files from the specified
 *    superblock (ie all the file names belonging to that fs).
 * RESULT
 *    NONE
 * NOTES
 *    - This is exclusively to debug the file open problem.
 * BUGS
 * TODO
 */
void mfs_dump_all_fs_files(struct super_block *sb);

/* NAME
 *    mfs_clear_migrated_flag
 * DESCRIPTION
 *    Clears the migrated flag.
 * RESULT
 *    0 for succeeded, else errno.
 * NOTES
 * BUGS
 * TODO
 */
int mfs_clear_migrated_flag(mfs_dentry_t *dentry);

/* NAME
 *    mfs_set_migrated_flag
 * DESCRIPTION
 *    Sets the migrated flag.
 * RESULT
 *    0 for succeeded, else errno.
 * NOTES
 * BUGS
 * TODO
 */
int mfs_set_migrated_flag(mfs_dentry_t *dentry);

#endif /* !_MFS_COM_INTERFACE_H */





