/* Project mfs --- Migration file system
 * mfs_interface.h -- User/kernel communication definitions
 * Creation: 17/10/1999
 * Update:   27/02/2000
 * V1.0 PV004
 * AUTHOR
 *    Marc SCHAEFER <schaefer@alphanet.ch>
 * LICENSE
 *    GPL
 * DESCRIPTION
 *    This include file defines data types used in the mfsd/mfs communication.
 * NOTES
 * BUGS
 * TODO
 * BASED-ON
 *    Original work.
 * MODIFICATION-HISTORY
 *    17/10/99  schaefer  Created this file.
 *    24/02/00  schaefer  Added NOTIFY_LIVES.
 *    27/02/00  schaefer  Ugly workaround for two FIST bugs.
 *    27/02/00  schaefer  NOTIFY_LIVES returns mfs_info_t struct.
 * $Id: mfs_interface.h,v 1.21 2000/02/27 18:12:15 schaefer Exp $
 */

#ifndef _MFS_INTERFACE_H
#define _MFS_INTERFACE_H

/* IOCTL definitions */

/* NOTES
 *    - It is important, when using a structure as parameter, that
 *      all the types used in the structure are the same in kernel
 *      driver and in user-space utilities. That's why we
 *      define our own types.
 *    - See also mfs_user_ioctls.h
 *    - These ioctls are done through the MFS_SPECIAL_MFSD_FILE, which
 *      must exist and is handled specially (open rw).
 * BUGS
 *    - We should allocate those so that no other fs ioctl ressembles ours.  
 */
#define MFS_IOCTL_GET_REQUEST_BLOCK   _IOR(0x15, 1, mfs_request_t)
#define MFS_IOCTL_SEND_ANSWER_BLOCK   _IOW(0x15, 2, mfs_answer_t)
#define MFS_IOCTL_ATOMIC_MIGRATE      _IOW(0x15, 3, mfs_atomic_migrate_t)

/* This one exists just to allow the first ioctl() to still go through
 * select(). It registers the file and mfsd process. Exactly with
 * the initial GET_REQUEST_BLOCK if no NOTIFY_LIVES, mfs will notice
 * when the file is closed and cancel ongoing requests. Returns
 * a mfs_info_t struct.
 */
#define MFS_IOCTL_NOTIFY_LIVES        _IOW(0x15, 4, mfs_info_t)

/* Special file name for mfsd ioctls */

#define MFS_SPECIAL_MFSD_FILE "fake_mfsd_file"

/* Type definitions */

typedef enum {
   MFS_COMMAND_GET_INDEX_OBJECT,
      /* Get the requested object from index, if any, and wake me up. */
   MFS_COMMAND_SYSTEM_LOW_IN_SPACE,
      /* System is low in space. Migrate upto low threshold and wake me up. */
   MFS_COMMAND_UNMIGRATE_OBJECT,
      /* System needs the listed file unmigrated. Wake me up after. */
   MFS_COMMAND_SHUTDOWN,
      /* mfsd is requested to shutdown itself (e.g. for umount().
       * In that case, mfsd should NOT reply but instead close the file.
       * This allows to handle a killed mfsd gracefully, too.
       */
   MFS_COMMAND_CREATE_MIGRATED_OBJECT,
      /* Add a new reference to a migrated object (indirect from
       * mfs_import) */
   MFS_COMMAND_DELETE_INDEX_OBJECT,
      /* System has invalidated an index object. Wake me up after
       * deleting references to this object, but do not delete
       * the object itself.
       */
   MFS_COMMAND_MIGRATE_OBJECT,
      /* Migrate the named file and wake me up. */
   MFS_COMMAND_WORK_AROUND
      /* Used by unlink() to free the space on the fs, working around
       * a FIST bug.
       */
} mfs_command_t;

typedef void *mfs_private_t;

/* BUGS
 *    - Should be somehow merged with mfs_user_ioctls.h
 */
typedef struct {
   mfs_import_object_locator_t locator; /* private to unmigration method */

   mfs_index_object_size_t size;
   mfs_uid_t uid;
   mfs_gid_t gid;
   mfs_time_t mtime; /* @@@ others? @@@ */
   mfs_mode_t mode;

   char migrated_on[MAX_MIGRATED_ON_PATHNAME_LENGTH];   
} mfs_import_data_t;

/* MFS_IOCTL_GET_REQUEST_BLOCK */
typedef struct {
   mfs_command_t command;
   mfs_private_t private;

   char path_name[MAX_PATHNAME_LENGTH];

   mfs_import_data_t import; /* only for MFS_COMMAND_CREATE_MIGRATED_OBJECT */
} mfs_request_t;

typedef enum {
   MFS_STATUS_OK,
   MFS_STATUS_OBJECT_NOT_FOUND,
   MFS_STATUS_OBJECT_EXISTS,
   MFS_STATUS_MFSD_UNAVAILABLE,
   MFS_STATUS_IO_ERROR,
   MFS_STATUS_NO_FREED_SPACE
} mfs_status_t;

/* MFS_IOCTL_SEND_ANSWER_BLOCK */
typedef struct {
   mfs_private_t private;
   mfs_status_t status;

   mfs_index_object_size_t size; /* size of file, if GET_INDEX_OBJECT */
} mfs_answer_t;

/* This special request is the parameter to the MFS_IOCTL_ATOMIC_MIGRATE
 * request.
 */
typedef struct {
   char path_name[MAX_PATHNAME_LENGTH];
} mfs_atomic_migrate_t;

#define MFS_PROTOCOL_VERSION 1

typedef struct {
   int protocol_version; /* MFS_PROTOCOL_VERSION */

   int queue_depth; /* Maximum number of parallel requests. Additionnal
                     * are put to sleep. This should be matched to
                     * the number of requests that mfsd can handle
                     * simultaneously.
                     */
   int reserved_entries; /* Number of entries reserved for index queries. */
} mfs_info_t;

#endif /* !_MFS_INTERFACE_H */




