/* Project mfs --- Migration file system
 * main.c -- mfs_import main module
 * Creation: 11/12/1999
 * Update:   11/12/1999
 * V1.0 PV001
 * AUTHOR
 *    Marc SCHAEFER <schaefer@alphanet.ch>
 * LICENSE
 *    GPL
 * DESCRIPTION
 *    This implements the command line parsing and initialization for the
 *    mfs_import tool. The mfs_import tool allows to incorporate an
 *    existing tar file into a mfs filesystem (as migrated data).
 *    The mfs filesystem must be mounted, and ioctl()s will be used
 *    to communicate the information to the mfsd process through the mfs
 *    filesystem.
 * NOTES
 *    - The first version will not be very optimized in the meaning that
 *      we will only tell mfs the size and the name of the file, and
 *      the name of the tar file. Additional information such as
 *      optimization information (block number in tar file) is not
 *      given by this version.
 *    - In the tar file name would be encoded the file number and
 *      the autochanger drive name, e.g. with # or something.
 *    - Exit codes are as follows:
 *         0 completed successfully.
 *         1 completed, but not all files were referenced without error.
 *         2 bad mount point.
 *         3 incorrect parameters.
 *    - Usage is, for example:
 *         dd if=/dev/nst0 bs=32k | tar tvf - | mfs_import /mnt tape-3-file-1
 * BUGS
 * TODO
 * BASED-ON
 *    Original work.
 * MODIFICATION-HISTORY
 *    11/12/99  schaefer  Created this file.
 * $Id: main.c,v 1.4 1999/12/19 16:56:24 schaefer Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <pwd.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include "../mfs/mfs_definitions.h"
#include "../mfs/mfs_user_ioctls.h"

/* Type definitions */

#define EXIT_ERROR_INVALID_PARAMETERS 3
#define EXIT_ERROR_BAD_MOUNT_POINT 2
#define EXIT_ERROR_NOT_ALL_OK 1
#define EXIT_ERROR_NONE 0

/* Private function definitions */

/* NAME
 *    usage
 * DESCRIPTION
 *    Prints a short usage summary on stderr.
 * RESULT
 *    NONE
 * NOTES
 * BUGS
 * TODO
 */
static void usage(char *program_name);

/* NAME
 *    do_import
 * DESCRIPTION
 *    Imports all the files in the tar file list from stdin into the
 *    mfs filesystem at the current directory.
 * RESULT
 *    EXIT_ERROR_NOT_ALL_OK or EXIT_ERROR_NONE
 * NOTES
 * BUGS
 * TODO
 */
static int do_import(fd_t fd, char *coming_from);

/* NAME
 *    create_file_as_migrated
 * DESCRIPTION
 *    Converts the passed parameters into valid ioctl() parameters,
 *    and calls the ioctl() which will atomically create the file,
 *    the index entry, and set the file as migrated.
 * RESULT
 *    TRUE if succeeded
 * NOTES
 * BUGS
 * TODO
 */
static BOOL create_file_as_migrated(char *name,
                                    char *coming_from,
                                    mfs_uid_t uid,
                                    mfs_gid_t gid,
                                    mfs_mode_t mode,
                                    mfs_time_t mtime,
                                    mfs_index_object_size_t size,
                                    fd_t fd);

/* NAME
 *    create_directory_path
 * DESCRIPTION
 *    Does the equivalent of a ``mkdir -p'' in a faster/less dangerous
 *    way.
 * RESULT
 *    TRUE if succeeded
 * NOTES
 * BUGS
 *    - should also have user, group, etc.
 * TODO
 */
static BOOL create_directory_path(char *name);

/* NAME
 *    decode_strings
 * DESCRIPTION
 *    Decode strings into values.
 * RESULT
 *    TRUE if succeeded
 * NOTES
 * BUGS
 * TODO
 */
static BOOL decode_strings(char **strings,
                           mfs_mode_t *mode,
                           mfs_uid_t *uid,
                           mfs_gid_t *gid,
                           mfs_index_object_size_t *size,
                           mfs_time_t *mtime);

/* Function implementation */

int main(int argc, char **argv) {
   int result = EXIT_ERROR_INVALID_PARAMETERS;

   if (argc == 3) {
      if (!chdir(argv[1])) {
         fd_t f = open(argv[1], O_RDONLY);
         /* @@@ or should we do it from the DIR for each file ? @@@ */

         if (f != -1) {
            result = do_import(f, argv[2]);
            close(f);
         }
         else {
            perror("open()");
            result = EXIT_ERROR_BAD_MOUNT_POINT;
         }
      }
      else {
         perror("chdir()");
         result = EXIT_ERROR_BAD_MOUNT_POINT;
      }
   }

   if (result ==  EXIT_ERROR_INVALID_PARAMETERS) {
      usage(argv[0]);
   }

   exit(result);
}

static void usage(char *program_name) {
   fprintf(stderr,
           "%s mfs-mount-point coming-from-name < tar-list\n",
           program_name);
   fprintf(stderr, "%s: bad args.\n", program_name);
}

static int do_import(fd_t fd, char *coming_from) {
   char buffer[MAX_PATHNAME_LENGTH];
   int result = EXIT_ERROR_NONE;

   while (fgets(buffer, sizeof(buffer), stdin)) {
      int length = strlen(buffer);
      char *strings[7]; /* date and hour */
      int i;
      mfs_index_object_size_t size;
      mfs_uid_t uid;
      mfs_gid_t gid;
      mfs_mode_t mode;
      mfs_time_t mtime;
      char *next;
      char *current;

      if (buffer[length - 1] != '\n') {
         fprintf(stderr, "Presumably truncated line %s. Aborting.\n", buffer);
         result = EXIT_ERROR_NOT_ALL_OK;
         break;
      }

      buffer[length - 1] = '\0';

      next = buffer;
      current = strsep(&next, " "); /* next token */
      i = 0;
      while (i < (sizeof(strings) / sizeof(char *))) {
         if (current) {
            if (!current || (*current != '\0')) {
               strings[i] = current;
#ifdef DEBUG
               printf("%d: %s.\n", i, current);
#endif /* DEBUG */
               i++;
            }
            current = strsep(&next, (i == 1) ? "/" : " ");
         }
         else {
            result = EXIT_ERROR_NOT_ALL_OK;
            break;
         }
      }

      if (result != EXIT_ERROR_NONE) {
         fprintf(stderr, "Presumably bad line %s. Aborting.\n", buffer);
         result = EXIT_ERROR_NOT_ALL_OK;
         break;
      }

      if (!decode_strings(strings,
                          &mode, &uid, &gid, &size, &mtime)) {
         fprintf(stderr,
                 "Can't decode fields. Aborting\n");
         result = EXIT_ERROR_NOT_ALL_OK;
         break;
      }

      if (strings[0][0] == 'd') {
         if (!create_directory_path(strings[6])) {
            fprintf(stderr,
                    "Can't create directory path %s. Aborting.\n",
                     strings[6]);
            result = EXIT_ERROR_NOT_ALL_OK;
            break;
         }
      }
      else if (strings[0][0] == '-') {
         if (!create_file_as_migrated(strings[6],
                                      coming_from,
                                      uid,
                                      gid,
                                      mode,
                                      mtime,
                                      size,
                                      fd)) {
            fprintf(stderr,
                    "Can't create file %s as migrated. Aborting.\n",
                     strings[6]);
            result = EXIT_ERROR_NOT_ALL_OK;
            break;
         }
      }
      else {
         fprintf(stderr,
                 "Ignored file %s type %d.\n",
                 strings[6],
                 strings[0][0]);
      }
   }

   return result;
}

static BOOL create_directory_path(char *name) {
   char buffer[MAX_PATHNAME_LENGTH];
   int size = strlen(name);

   if (size >= sizeof(buffer)) {
      fprintf(stderr, "cdp: size too big.\n");
      return FALSE;
   }

   while (size) {
      char *tmp;

      while (size && (name[size - 1] == '/')) {
         size--;
      }

      strncpy(buffer, name, size);
      buffer[size] = '\0';

#ifdef DEBUG
      fprintf(stderr, "mkdir(%s)\n", buffer);
#endif /* DEBUG */
      if (mkdir(buffer, 0777 /* @@@ */)) {
         if (errno == EEXIST) {
            return TRUE;
         }
         else {
            perror("mkdir()");
            return FALSE;
         }
      }

      tmp = strrchr(buffer, '/');
      if (tmp && (tmp - buffer)) {
         size = (tmp - buffer);
      }
      else {
         size = 0;
      }
   }
   
   return TRUE;
}

static BOOL create_file_as_migrated(char *name,
                                    char *coming_from,
                                    mfs_uid_t uid,
                                    mfs_gid_t gid,
                                    mfs_mode_t mode,
                                    mfs_time_t mtime,
                                    mfs_index_object_size_t size,
                                    fd_t fd) {
   mfs_import_t import;

#ifdef DEBUG
   printf("create_file_as_migrated(%s, %s, %d, %d, %d, %d, %d)\n",
          name, coming_from, uid, gid, mode, mtime, size);
#endif /* DEBUG */

   import.locator = 0; /* @@@ for optimization, private to me @@@ */
   import.size = size;
   import.uid = uid;
   import.gid = gid;
   import.mode = mode;
   import.mtime = mtime;

   if (strlen(name) >= sizeof(import.file_name)) {
      fprintf(stderr, "%s: too long filename.\n", name);
      return FALSE;
   }

   if (strlen(coming_from) >= sizeof(import.migrated_on)) {
      fprintf(stderr, "%s: too long filename.\n", name);
      return FALSE;
   }

   strncpy(import.file_name, name, sizeof(import.file_name));
   strncpy(import.migrated_on, coming_from, sizeof(import.migrated_on));

   if (ioctl(fd, MFS_IOCTL_IMPORT_OBJECT, &import)) {
      perror("ioctl(MFS_IOCTL_IMPORT_OBJECT)");
      return FALSE;
   }

   return TRUE;
}

static BOOL decode_strings(char **strings,
                           mfs_mode_t *mode,
                           mfs_uid_t *uid,
                           mfs_gid_t *gid,
                           mfs_index_object_size_t *size,
                           mfs_time_t *mtime) {
   BOOL result = TRUE;
   char *tmp;
   int j;
   struct passwd *pw;

   /* BUGS
    *    - Could be made better.
    */
   *mode = 0;
   tmp = strings[0] + 1;
   for (j = 0; j < 3; j++) {
      int shift = ((2 - j) * 3);

      if (*tmp == 'r') {
         *mode |= (4 << shift);
      }
      tmp++;

      if (*tmp == 'w') {
         *mode |= (2 << shift);
      }
      tmp++;

      if ((*tmp == 'x') || (*tmp == 's') || (*tmp == 'x')) {
         *mode |= (1 << shift);
      }
      if ((*tmp == 't') || (*tmp == 'T')) {
         *mode |= 1 << 9;
      }
      else if ((*tmp == 's') || (*tmp == 'S')) {
         if (j == 0) {
            *mode |= 1 << 11;
         }
         else if (j == 1) {
            *mode |= 1 << 10;
         }
      }
      tmp++;
   } 

   *size = strtol(strings[3], &tmp, 10);
   if ((strings[3][0] == '\0') || (*tmp != 0)) {
      fprintf(stderr,
              "Can't get size for %s. Aborting\n", strings[3]);
      result = FALSE;
   }

   pw = getpwnam(strings[1]);
   if (pw) {
      *uid = pw->pw_uid;
#if 0 /* crashes */
      free(pw); /* @@@ needed? @@@ */
#endif
      *gid = 0; /* @@@ HOW TO GET THIS ??? getgrpgid() ?? @@@ */
   }
   else {
      fprintf(stderr,
              "WARNING: Unknown UID %s. Setting to 0/0, mode 0.\n",
              strings[1]);
      *mode = 0;
      *uid = 0;
      *gid = 0;
   }

   /* @@@ how to transform a time in string to time_t ? @@@ */

   *mtime = 0; /* epoch @@@ */

   return result;
}
