/* Project mfs --- Migration file system
 * changer.c -- Changer implementation
 * Creation: 05/03/2000
 * Update:   05/03/2000
 * V1.0 PV001
 * AUTHOR
 *    Marc SCHAEFER <schaefer@alphanet.ch>
 * LICENSE
 *    GPL
 * DESCRIPTION
 *    This implements the changer subsystem, responsible for
 *    retrieving the correct tape in a chosen drive,
 *    and starting tape I/O.
 * NOTES
 *    - This might run in multitasking (e.g. if a changer has multiple
 *      drives).
 * BUGS
 *    - Should be reworked into:
 *         client part ---- server part ---- SCHEDULER ---- THIS
 *    - Missing interface specification.
 *    - Should be able to mount tapes and MOs, etc.
 * TODO
 * BASED-ON
 *    Original work.
 * MODIFICATION-HISTORY
 *    05/03/00  schaefer  Created this file.
 * $Id: tapectl.c,v 1.1 2000/03/04 13:28:13 schaefer Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include "changer_interface.h"
#include "tape/tapectl_interface.h"

/* Private type definitions */

#define CHANGER_SLOT_UNKNOWN (-1)
typedef unsigned short changer_slot_t;
typedef int changer_fd_t;

typedef struct {
   changer_fd_t fd;

   changer_slot_t current;   

   /* @@@ drive array, slot array, for _slot_ functions @@@ 
} changer_info_t;

changer_descriptor_t changer_initialize(const char *changer_device_name) {
   changer_info_t *changer_info;

   changer = (changer_info_t *) malloc(sizeof(changer_info_t));
   if (changer) {
      changer->fd = 0; /* @@@ */
      changer->current = CHANGER_SLOT_UNKNOWN;
   }

   return changer_info;
}

void changer_deinitialize(changer_descriptor_t *changer_descriptor) {
   free(*changer_descriptor);
   *changer_descriptor = NULL;
}

BOOL changer_tape_mount(changer_descriptor_t changer_descriptor,
                        const char *container_name,
                        tape_descriptor_t *tape_descriptor) {
   changer_fd_t initial;
   changer_fd_t slot;
   changer_info_t *changer_info = (changer_info_t *) changer_descriptor;
   BOOL result = FALSE;

   *tape_descriptor = NULL;

   /* NOTES
    *    - The idea is to first support a mounted tape with matching
    *      label. Then to try in the last access database, then to
    *      try all slots.
    *    - Except for the tape mounted case (ie tape_descriptor open)
    *      we re-read the label to verify.
    *    - In summary, if the tape is mounted, it is opened, and
    *      thus we can play with it without any problem (no need
    *      to verify label). If not, we quickly mount it. If the
    *      operator did shuffle the tapes, we will recover from this
    *      situation automatically (an AUDIT LIBRARY command is thus
    *      not needed).
    */

   initial = changer_info->current;
   if (initial == CHANGER_SLOT_UNKNOWN) {
      initial = 0; /* first slot */
   }

   slot = changer_slot_last_seen(changer_info,
                                 container_name,
                                 tape_descriptor);
   if (slot == CHANGER_SLOT_UNKNOWN) {
      slot = initial;
   }

   if (*tape_descriptor == TAPE_FD_ERROR) {
      /* Tape is not mounted, we need to search for it, starting at
       * the most probable location.
       */
      do {
         if (changer_slot_load(changer_info, tape_descriptor, slot)) {
            if (changer_slot_matches(changer_info,
                                     container_name,
                                     tape_descriptor,
                                     TRUE)) { /* @@@ unmounts @@@ */
               /* We found the correct tape */
               break; /* out of while */
            }
         }
	 if (slot < changer_info->slot_amount) {
	    slot++;
	 }
	 else {
	    slot = 0;
	 }
      } while (slot != initial);
   }
   else {
      result = TRUE; /* tape is already mounted */
   }

   return result;   
}

/* @@@ the slot functions maintain database of slots and also
 * of drives.
 */