/* Project mfs --- Migration file system
 * tapectl.c -- Generic tape control functions
 * Creation: 04/03/2000
 * Update:   04/03/2000
 * V1.0 PV001
 * AUTHOR
 *    Marc SCHAEFER <schaefer@alphanet.ch>
 * LICENSE
 *    GPL
 * DESCRIPTION
 *    This implements the basic tape I/O and movement functions. This also
 *    includes using the best capabilities of the drives.
 * NOTES
 * BUGS
 * TODO
 *    - This interface is really simplified compare to what
 *      tapefs's tape module can do. Maybe investigate what can
 *      be taken from tapefs without complexifying too much.
 *      Probably the best to do is to implement changer + tar
 *      migration/unmigration first and see what services we need.
 * BASED-ON
 *    Original work. However the idea of capabilities and the implementation
 *    of the tapectl_move() function is taken from my tapefs.
 * MODIFICATION-HISTORY
 *    04/03/00  schaefer  Created this file.
 * $Id: tapectl.c,v 1.1 2000/03/04 13:28:13 schaefer Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <sys/mtio.h>
#include <sys/ioctl.h>
#include "../../../mfs/mfs_definitions.h"
#include "tapectl_interface.h"

/* Private type definitions */

typedef int tape_fd_t;

#define TAPE_CAPABILITIES_ @@@ /* see tapefs! @@ */
typedef unsigned char tape_capabilities_t;
#define TAPE_DEFAULT_CAPABILITIES 0 /* @@@ */

typedef struct {
   tape_fd_t fd;

   tape_capabilities_t caps;
} tape_info_t;

/* Function definitions */

tape_descriptor_t tapectl_open(const char *device_name, BOOL readwrite) {
   tape_info_t *tape_info;

   tape_info = malloc(sizeof(tape_info_t));
   if (tape_info) {
      tape_info->caps = TAPE_DEFAULT_CAPABILITIES;
      tape_info->fd = open(device_name,
                           readwrite ? O_RDWR : O_RDONLY);
      if (tape_info->fd == -1) {
         fprintf(stderr, "tapectl: open(%s, %s) failed: %s.\n",
                         device_name,
                         readwrite ? "RW" : "R",
                         strerror(errno));
         free(tape_info);
         tape_info = NULL;
      }
   }
   else {
      fprintf(stderr, "tapectl: malloc() failed.\n");
   }

   return (tape_descriptor_t) tape_info;
}

BOOL tapectl_close(tape_descriptor_t *tape_fd) {
   BOOL result = TRUE;
   tape_info_t *tape_info = (tape_info_t *) *tape_fd;

   if (close(tape_info->fd)) {
      fprintf(stderr, "tapectl: close() failed: %s.\n",
                      strerror(errno));
      result = FALSE;
   }

   free(tape_info);

   *tape_fd = NULL;

   return result;
}

BOOL tapectl_get_block_size(tape_descriptor_t tape_fd,
                            tape_block_size_t *tape_block_size) {
   BOOL result = TRUE;
   tape_info_t *tape_info = (tape_info_t *) tape_fd;
   struct mtget status;

   if (ioctl(tape_info->fd, MTIOCGET, &status)) {
      fprintf(stderr, "tapectl: ioctl(MTIOCGET) failed: %s.\n",
                      strerror(errno));
      result = FALSE;
   }
   else {
      *tape_block_size
         = (status.mt_dsreg >> MT_ST_BLKSIZE_SHIFT) & MT_ST_BLKSIZE_MASK;
   }

   return result;
}

BOOL tapectl_move(tape_descriptor_t tape_fd,
                  tape_file_t file,
                  tape_record_t record) {
   BOOL result = TRUE;
   tape_info_t *tape_info = (tape_info_t *) tape_fd;
   struct mtop op;

/* @@@ look at tapefs implementation */

   if (ioctl(tape_info->fd, MTIOCTOP, &op)) {
      fprintf(stderr, "tapectl: ioctl(MTIOCTOP) failed: %s.\n",
                      strerror(errno));
      result = FALSE;
   }

   return result;
}

BOOL tapectl_read(tape_descriptor_t tape_fd,
                  unsigned char *buffer,
                  tape_io_size_t *size) {
   tape_info_t *tape_info = (tape_info_t *) tape_fd;
   ssize_t length;

   length = read(tape_info->fd, buffer, *size);
   if (length == -1) {
      fprintf(stderr, "tapectl: read() failed: %s.\n",
                      strerror(errno));
      *size = 0;
   }
   else {
      *size = length;
   }

   return (length != -1);
}

BOOL tapectl_write(tape_descriptor_t tape_fd,
                   const unsigned char *buffer,
                   tape_io_size_t *size) {
   tape_info_t *tape_info = (tape_info_t *) tape_fd;
   ssize_t length;

   length = write(tape_info->fd, buffer, *size);
   if (length == - 1) {
      fprintf(stderr, "tapectl: write() failed: %s.\n",
                      strerror(errno));
      *size = 0;
   }
   else {
      *size = length;
   }

   return (length != -1);
}

BOOL tapectl_write_filemark(tape_descriptor_t tape_fd) {
   BOOL result = TRUE;
   tape_info_t *tape_info = (tape_info_t *) tape_fd;
   struct mtop op;

   op.mt_op = MTWEOF;
   op.mt_count = 1;
   if (ioctl(tape_info->fd, MTIOCTOP, &op)) {
      fprintf(stderr, "tapectl: ioctl(MTIOCTOP, MTWEOF) failed: %s.\n",
                      strerror(errno));
      result = FALSE;
   }

   return result;
}

BOOL tapectl_get_position(tape_descriptor_t tape_fd,
                          tape_file_t *file,
                          tape_record_t *record) {
   BOOL result = TRUE;
   tape_info_t *tape_info = (tape_info_t *) tape_fd;
   struct mtget status;

   /* @@@ if tape drive supports SEEK, then we should use
    * MTIOCPOS.
    */

   if (ioctl(tape_info->fd, MTIOCGET, &status)) {
      fprintf(stderr, "tapectl: ioctl(MTIOCGET) failed: %s.\n",
                      strerror(errno));
      result = FALSE;
   }
   else {
      *file = status.mt_fileno;
      *record = status.mt_blkno;
   }

   return result;
}
