/* Project mfs --- Migration file system
 * tapectl_interface.h -- Generic tape control interface definitions
 * Creation: 03/03/2000
 * Update:   03/03/2000
 * V1.0 PV001
 * AUTHOR
 *    Marc SCHAEFER <schaefer@alphanet.ch>
 * LICENSE
 *    GPL
 * DESCRIPTION
 *    This include file defines the data types and functions used for
 *    control of a tape device.
 * NOTES
 *    - In case of errors, perror(), errno or strerror() can be used.
 *    - We assume non variable block mode (much easier).
 * BUGS
 * TODO
 *    - This interface is really simplified compare to what
 *      tapefs's tape module can do. Maybe investigate what can
 *      be taken from tapefs without complexifying too much.
 *      Probably the best to do is to implement changer + tar
 *      migration/unmigration first and see what services we need.
 * BASED-ON
 *    Original work.
 * MODIFICATION-HISTORY
 *    03/03/00  schaefer  Created this file.
 * $Id: tapectl_interface.h,v 1.2 2000/03/04 13:28:14 schaefer Exp $
 */

#ifndef _TAPECTL_INTERFACE_H
#define _TAPECTL_INTERFACE_H

/* Type definitions */

#define TAPE_FD_ERROR NULL
typedef void * tape_descriptor_t;

typedef unsigned long int tape_file_t;
typedef unsigned long int tape_record_t; /* Max. 2 TB @ 512 bytes/block
                                          * per tape file.
                                          */
typedef unsigned long int tape_block_size_t;
typedef unsigned long int tape_io_size_t;

/* Function definitions */

/* NAME
 *    tapectl_open
 * DESCRIPTION
 *    Opens a tape device with the specified mode.
 * RESULT
 *    a tape_descriptor_t, TAPE_FD_ERROR in case of open error.
 * NOTES
 * BUGS
 * TODO
 */
tape_descriptor_t tapectl_open(const char *device_name, BOOL readwrite);

/* NAME
 *    tapectl_close
 * DESCRIPTION
 *    Closes a tape device. The associated tape_fd should no longer
 *    be used, including if the close was unsuccessful.
 * RESULT
 *    TRUE if operation successful.
 * NOTES
 * BUGS
 * TODO
 */
BOOL tapectl_close(tape_descriptor_t *tape_fd);

/* NAME
 *    tapectl_get_block_size
 * DESCRIPTION
 *    Returns the currently used block size.
 * RESULT
 *    TRUE if operation successful.
 * NOTES
 *    - Zero means variable block size. In that case, the
 *      user must do its computations itself, knowing the
 *      I/O size used in preceeding tapefd_write()s.
 * BUGS
 * TODO
 */
BOOL tapectl_get_block_size(tape_descriptor_t tape_fd,
                            tape_block_size_t *tape_block_size);

/* NAME
 *    tape_move
 * DESCRIPTION
 *    Moves the tape to the specified absolute position.
 * RESULT
 *    TRUE if operation successful.
 * NOTES
 *    - tapectl_move(0, 0, 0) == rewind(). This should be done
 *      at tape opening, for example.
 * BUGS
 * TODO
 */
BOOL tapectl_move(tape_descriptor_t tape_fd,
                  tape_file_t file,
                  tape_record_t record);

/* NAME
 *    tapectl_read
 * DESCRIPTION
 *    Reads data from the tape.
 * RESULT
 *    TRUE if operation successful.
 * NOTES
 *    - Must read a multiple of block size.
 *    - size will reflect, on exit, the actual read data (smaller
 *      than requested can happen with result TRUE).
 * BUGS
 * TODO
 */
BOOL tapectl_read(tape_descriptor_t tape_fd,
                  unsigned char *buffer,
                  tape_io_size_t *size);

/* NAME
 *    tapectl_write
 * DESCRIPTION
 *    Writes data to the tape.
 * RESULT
 *    TRUE if operation successful.
 * NOTES
 *    - Must write a multiple of block size.
 *    - size will reflect, on exit, the actual read data (smaller
 *      than requested can happen with result TRUE).
 * BUGS
 * TODO
 */
BOOL tapectl_write(tape_descriptor_t tape_fd,
                   const unsigned char *buffer,
                   tape_io_size_t *size);

/* NAME
 *    tapectl_write_filemark
 * DESCRIPTION
 *    Write a file mark on the tape.
 * RESULT
 *    TRUE if operation successful.
 * NOTES
 * BUGS
 * TODO
 */
BOOL tapectl_write_filemark(tape_descriptor_t tape_fd);

/* NAME
 *    tapectl_get_position
 * DESCRIPTION
 *    Returns the current absolute position.
 * RESULT
 *    TRUE if operation successful.
 * NOTES
 * BUGS
 * TODO
 */
BOOL tapectl_get_position(tape_descriptor_t tape_fd,
                          tape_file_t *file,
                          tape_record_t *record);
                  
#endif /* !_TAPECTL_INTERFACE_H */
