/* dcache.c -- Directory cache implementation.
 * Marc SCHAEFER <schaefer@alphanet.ch>
 * Creation:    25/08/96
 * Last update: 26/10/96
 * V1.0 PV003
 * NOTES
 *    - We could have used a dbm-like association.
 *    - No compaction is needed because we recreate the file at end.
 *    - Or what about a tree-like structure ressembling directories ?
 *    - Or what about using cpio archive which could then be imported
 *      easily. Well, we could have independance. This is anyway only
 *      the interface hiding the complexity: we can imagine someone
 *      mounting a tar or a cpio based on the info from here.
 * BUGS
 *    - bad definition of namelen et al
 *    - bad manual for fread: returning number of ``blocks'', not char.
 *    - non implementation independance: saving binary structures.
 *    - using hardcoded constants.
 *    - missing spec for private functions
 * TODO
 *    - dcache_stat, dcache_create, dcache_remove using the dcache_internal
 *      functions make_node, lookup, delete.
 *    - test
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <sys/time.h>
#include <unistd.h>
#include <memory.h>
#include <malloc.h>
#include <string.h>
#include "types.h"
#include "dcache_interface.h"
#include "dcache_private.h"

/* Type definitions */

/* Round up to even value */
#define DYNAMIC_LEN(x) ((((x) + 1) >> 1) << 1)
#define DYNAMIC_NAMELEN(x) (DYNAMIC_LEN((x).description.namelen))
#define DYNAMIC_STRLEN(x) (DYNAMIC_LEN(strlen(x)))

typedef struct {
   char *pathname;
   dcache_node_t *first;
} dcache_internal_key_t;

/* Macros */

#ifndef DEBUG_CACHE
#define dcache_internal_cache_debug
#endif

/* Private function definitions */

static dcache_node_t *dcache_internal_cache_make_node(dcache_entry_t the_entry,
                                                      const char *pathname);
static BOOL dcache_internal_cache_sync(const char *pathname,
                                       dcache_node_t *first);
static void dcache_internal_free_all(dcache_node_t *first);
static dcache_node_t *dcache_internal_cache_lookup(const char *pathname,
                                                   dcache_node_t *first);
static void dcache_internal_cache_add(dcache_node_t *node,
                                      dcache_node_t **first);
static void dcache_internal_cache_delete(dcache_node_t *n,
                                         dcache_node_t **first);
static void dcache_internal_dump(dcache_node_t *n, char *pathname,
                                 int path_len, dcache_stat_t *buf);
#ifdef DEBUG_CACHE
static void dcache_internal_cache_debug(dcache_node_t *first);
#endif

/* Function implementation */

dcache_result_t dcache_make(char *pathname) {
   FILE *f;

   f = fopen(pathname, "w");
   if (f) {
      if (fclose(f) == 0) {
         return DCACHE_RESULT_OK;
      }
      else {
         MESSAGE(DEBUG_ERROR, "dcache.c", "dcache_make", "fclose failed",
                 pathname);
         return DCACHE_RESULT_ERROR_FATAL;
      }
      /* NOT REACHED */
   }
   else {
      MESSAGE(DEBUG_ERROR, "dcache.c", "dcache_make", "fopen failed",
                           pathname);
      return DCACHE_RESULT_ERROR_FATAL;
   }
   /* NOT REACHED */
}

/*
 * ASSUMPTION
 *    - binary order (little endian)
 *    - no compiler tricks
 * NOTES
 *    - sizeof(dcache_saved_entry_t *) intended
 */
dcache_key_t dcache_init(char *pathname) {
   dcache_internal_key_t *the_key = (dcache_internal_key_t *)
                                    malloc(sizeof(dcache_internal_key_t));
   BOOL result = TRUE;
   FILE *f;

   if (the_key) {
      the_key->first = NULL;
      the_key->pathname = strdup(pathname);
      if (the_key->pathname) {

	 f = fopen(pathname, "r");
	 if (f) {
	    dcache_entry_t the_entry;

	    while (result && (fread(&the_entry, sizeof(dcache_entry_t), 1,
				    f)
		   == 1)) {
	       dcache_node_t *node = (dcache_node_t *)
				     malloc(sizeof(dcache_node_t)
					    + DYNAMIC_NAMELEN(the_entry));

	       if (node) {
		  memcpy(&node->the_entry, &the_entry, sizeof(dcache_entry_t));
		  if (fread(((char *) &node->the_entry)
                            + sizeof(dcache_entry_t),
			    DYNAMIC_NAMELEN(the_entry),
			    1,
			    f)
		      == 1) {
		     dcache_internal_cache_add(node, &the_key->first);
		  }
		  else {
		     MESSAGE(DEBUG_FATAL, "dcache.c", "dcache_init",
					  "short read pathname",
                                          pathname);
		     result = FALSE;
		     free(node);
		  }
	       }
	       else {
		  MESSAGE(DEBUG_FATAL, "dcache.c", "dcache_init",
				       "can't malloc node", "");
		  result = FALSE;
	       }
	    }
	    /* error ignored */

	    fclose(f); /* same */
	 }
	 else {
	    MESSAGE(DEBUG_FATAL, "dcache.c", "dcache_init",
				 "can't open", pathname);
	    result = FALSE;
	 }
      }
      else {
	 MESSAGE(DEBUG_FATAL, "dcache.c", "dcache_init",
                              "can't strdup", pathname);
	 result = FALSE;
      }
   }
   else {
      MESSAGE(DEBUG_FATAL, "dcache.c", "dcache_init",
			   "can't malloc", "key");
      result = FALSE;
   }

   if (!result && the_key) { /* which is redundant */
      dcache_internal_free_all(the_key->first);
      free(the_key->pathname);
      free(the_key);
      the_key = NULL; /* new */
   }
   return (dcache_key_t) the_key;
}

dcache_result_t dcache_deinit(dcache_key_t key) {
   dcache_internal_key_t *the_key = (dcache_internal_key_t *) key;

   if (key) {
      if (dcache_internal_cache_sync(the_key->pathname, the_key->first)) {
	 dcache_internal_free_all(the_key->first);
	 free(the_key->pathname);
	 free(the_key);
	 return DCACHE_RESULT_OK;
      }
      else {
	 dcache_internal_free_all(the_key->first);
	 free(the_key->pathname);
	 free(the_key);
	 return DCACHE_RESULT_ERROR_FATAL;
      }
      /* NOT REACHED */
   }
   else {
      MESSAGE(DEBUG_FATAL, "dcache.c", "dcache_deinit",
                           "called with NULL key", "");
      return DCACHE_RESULT_ERROR_FATAL;
   }
   /* NOT REACHED */
}

dcache_result_t dcache_stat(dcache_key_t key,
                            const char *pathname,
                            dcache_stat_t *buf) {
   dcache_internal_key_t *the_key = (dcache_internal_key_t *) key;

   if (key) {
      dcache_node_t *the_node
         = dcache_internal_cache_lookup(pathname, the_key->first);

      if (the_node) {
         memcpy(buf, &(the_node->the_entry.description.descr),
                sizeof(dcache_stat_t));
         return DCACHE_RESULT_OK;
      }
      else {
         return DCACHE_RESULT_ERROR_NO_FILE;
      }
      /* NOT REACHED */
   }
   else {
      MESSAGE(DEBUG_FATAL, "dcache.c", "dcache_stat",
                           "called with NULL key", "");
      return DCACHE_RESULT_ERROR_FATAL;
   }
   /* NOT REACHED */
}

dcache_result_t dcache_create(dcache_key_t key,
                              const char *pathname,
                              dcache_stat_t buf) {
   dcache_internal_key_t *the_key = (dcache_internal_key_t *) key;

   if (key) {
      dcache_node_t *the_node
         = dcache_internal_cache_lookup(pathname, the_key->first);

      if (the_node) {
	 MESSAGE(DEBUG_WARN, "dcache.c", "dcache_create",
			     "existing", pathname);
         return DCACHE_RESULT_ERROR_EXISTING;
      }
      else {
         dcache_entry_t the_entry; /* without pathname */

         the_entry.free = FALSE;
         the_entry.description.magic = DCACHE_SAVED_ENTRY_T_MAGIC;
         memcpy(&(the_entry.description.descr), &buf, sizeof(dcache_stat_t));

         the_node = dcache_internal_cache_make_node(the_entry, pathname);

	 if (the_node) {
	    dcache_internal_cache_add(the_node, &the_key->first);
	    return DCACHE_RESULT_OK;
	 }
         else {
	    MESSAGE(DEBUG_ERROR, "dcache.c", "dcache_create",
				 "can't add", "");
            return DCACHE_RESULT_CANT_ADD;
         }
         /* NOT REACHED */
      }
      /* NOT REACHED */
   }
   else {
      MESSAGE(DEBUG_FATAL, "dcache.c", "dcache_create",
                           "called with NULL key", "");
      return DCACHE_RESULT_ERROR_FATAL;
   }
   /* NOT REACHED */
}

dcache_result_t dcache_remove(dcache_key_t key,
                              const char *pathname) {
   dcache_internal_key_t *the_key = (dcache_internal_key_t *) key;

   if (key) {
      dcache_node_t *the_node
         = dcache_internal_cache_lookup(pathname, the_key->first);

      if (the_node) {
         dcache_internal_cache_delete(the_node, &the_key->first);
         return DCACHE_RESULT_OK;
      }
      else {
         return DCACHE_RESULT_ERROR_NO_FILE;
      }
      /* NOT REACHED */
   }
   else {
      MESSAGE(DEBUG_FATAL, "dcache.c", "dcache_remove",
                           "called with NULL key", "");
      return DCACHE_RESULT_ERROR_FATAL;
   }
   /* NOT REACHED */
}

dcache_result_t dcache_sync(dcache_key_t key) {
   dcache_internal_key_t *the_key = (dcache_internal_key_t *) key;

   if (key) {
      if (dcache_internal_cache_sync(the_key->pathname, the_key->first)) {
         return DCACHE_RESULT_OK;
      }
      else {
         return DCACHE_RESULT_ERROR_FATAL;
      }
   }
   else {
      MESSAGE(DEBUG_FATAL, "dcache.c", "dcache_sync",
                           "called with NULL key", "");
      return DCACHE_RESULT_ERROR_FATAL;
   }
   /* NOT REACHED */
}

dcache_result_t dcache_browse(dcache_key_t key,
                              BOOL first,
                              dcache_browse_t *browsing_key,
                              dcache_stat_t *stat_buf,
                              char *pathname,
                              int path_len) {
   dcache_internal_key_t *the_key = (dcache_internal_key_t *) key;

   if (key) {
      dcache_result_t result = DCACHE_RESULT_OK;
      dcache_node_t *current;

      if (first) {
	 current = the_key->first;
      }
      else {
         current = (dcache_node_t *) *browsing_key;
      }

      if (current) {
         dcache_internal_dump(current, pathname, path_len, stat_buf);

         current = current->next;
      }
      else {
         result = DCACHE_RESULT_ERROR_NO_FILE;
      }

      *browsing_key = (void *) current;

      return result;
   }
   else {
      MESSAGE(DEBUG_FATAL, "dcache.c", "dcache_browse",
                           "called with NULL key", "");
      return DCACHE_RESULT_ERROR_FATAL;
   }
   /* NOT REACHED */
}

/* Private function implementation */

static dcache_node_t *dcache_internal_cache_make_node(dcache_entry_t the_entry,
                                                      const char *pathname) {
   size_t len = sizeof(dcache_node_t) + DYNAMIC_STRLEN(pathname) + 1;
   dcache_node_t *n;

   n = (dcache_node_t *) malloc(len);
   if (n) {
      memcpy((char *) &n->the_entry,
             (char *) &the_entry,
             sizeof(dcache_entry_t));
      n->the_entry.description.namelen = strlen(pathname) + 1;
      strcpy(((char *) &n->the_entry) + sizeof(dcache_entry_t),
             pathname);
   }
   return n;
}


static BOOL dcache_internal_cache_sync(const char *pathname,
                                       dcache_node_t *first) {
   BOOL result = TRUE;
   dcache_node_t *current = first;
   FILE *f;

   f = fopen(pathname, "w");
   if (f) {
      while (current) { /* some error checks missing */
         fwrite(&current->the_entry, sizeof(dcache_entry_t), 1, f);
         fwrite(((char *) &current->the_entry) + sizeof(dcache_entry_t),
                 DYNAMIC_NAMELEN(current->the_entry), 1, f);
	 current = current->next;
      }
      fclose(f);
   }
   return result;
}

static void dcache_internal_free_all(dcache_node_t *first) {
   dcache_node_t *next;

   while (first) {
      /* rien a faire sur le prev, car on efface tout */
      next = first->next;
      free(first); /* y compris dynamique */
      first = next;
   }
}

static dcache_node_t *dcache_internal_cache_lookup(const char *pathname,
                                                   dcache_node_t *first) {
   dcache_node_t *current = first;

   while (current && strcmp(pathname,
                            ((char *) &current->the_entry)
                            + sizeof(dcache_entry_t))) {
      current = current->next;
   }
   return current;
}

/* Without any tests if it exists already */
static void dcache_internal_cache_add(dcache_node_t *node,
                                      dcache_node_t **first) {
   node->next = *first;
   node->prev = NULL;
   if (*first) {
     (*first)->prev = node;
   }
   *first = node; 
   MESSAGE(DEBUG_INFO, "dcache.c", "dcache_internal_cache_add",
                       "added one node",
                       ((char *) &node->the_entry) + sizeof(node->the_entry));
}

static void dcache_internal_cache_delete(dcache_node_t *n,
                                         dcache_node_t **first) {
   if (n) {
      if (*first == n) {
         *first = n->next;
      }
      if (n->next) {
         n->next->prev = n->prev;
      }
      if (n->prev) {
         n->prev->next = n->next;
      }
      free(n);
   }
}

static void dcache_internal_dump(dcache_node_t *n, char *pathname,
                                 int path_len, dcache_stat_t *buf) {
   if (n) {
      int actual_size = MIN((path_len - 1),
                            (n->the_entry.description.namelen - 1)
                           );
      strncpy(pathname, ((char *) &n->the_entry) + sizeof(dcache_entry_t),
              actual_size);
      pathname[actual_size] = '\0';

      memcpy(buf, &(n->the_entry.description.descr), sizeof(dcache_stat_t));
   }
}

#ifdef DEBUG_CACHE
static void dcache_internal_cache_debug(dcache_node_t *first) {
   dcache_node_t *current = first;

   printf("DEBUG_CACHE\n");

   while (current) {
      printf("entry 0x%x\n", current);
      printf("   next: 0x%x\n", current->next);
      printf("   prev: 0x%x\n", current->prev);
      printf("   the_entry\n");
      printf("      free: %s\n", current->the_entry.free ? "TRUE" : "FALSE");
      printf("      where\n");
      printf("         tape: %ld\n", current->the_entry.where.tape);
      printf("         group: %ld\n", current->the_entry.where.group);
      printf("      description\n");
      printf("         magic: 0x%lx\n", current->the_entry.description.descr.magic);
      printf("         mode: 0x%x\n", current->the_entry.description.descr.mode);
      printf("         uid: %d\n", current->the_entry.description.descr.uid);
      printf("         gid: %d\n", current->the_entry.description.descr.gid);
      printf("         size: %ld\n", current->the_entry.description.descr.size);
      printf("         atime: %ld %s",
                       current->the_entry.description.descr.atime,
                       ctime(&current->the_entry.description.descr.atime));
      printf("         mtime: %ld %s",
                       current->the_entry.description.descr.mtime,
                       ctime(&current->the_entry.description.descr.mtime));
      printf("         ctime: %ld %s",
                       current->the_entry.description.descr.ctime,
                       ctime(&current->the_entry.description.descr.ctime));
      printf("         migration: %ld %s",
                       current->the_entry.description.descr.migration,
                       ctime(&current->the_entry.description.descr.migration));
      printf("         namelen: %ld\n", current->the_entry.description.descr.namelen);
      printf("         NAME: %s\n", ((char *) &current->the_entry) + sizeof(dcache_entry_t));
      current = current->next;
   }
}
#endif


