#ifndef DCACHE_INTERFACE_H
#define DCACHE_INTERFACE_H
/* dcache_interface.h -- Directory cache public interface definitions
 * Marc SCHAEFER <schaefer@alphanet.ch>
 * Creation:    25/08/96
 * Last update: 20/12/99
 * V1.0 PV003
 * DESCRIPTION
 *    This module implements a directory cache for fast locating files
 *    on a domain. Functions offered are:
 *       - registering of a file to the directory cache
 *       - removing of a reference in the directory cache
 *       - getting the structure of the directory cache
 *       - searching for a specific element.
 *    The module can handle more than one directory cache at the same
 *    time, because all information is passed with keys.
 * DEFINITIONS
 *    directory cache: file containing a summary of the headers for
 *                     all migrated files.
 *    domain:          a unified set of tapes.
 * NOTES
 * BUGS
 * TODO
 *    - missing functions for iteration, lists and so on. They are
 *      not needed as long as we do not need to check or recompact.
 *      Easier rebuild is simply to destroy and start over again.
 * MODIFICATION HISTORY
 *    20/12/99  schaefer  Adapted to mfs needs
 * $Id: dcache_interface.h,v 1.2 1999/12/20 15:26:13 schaefer Exp $
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <unistd.h>
#include "types.h"

#include "../../mfs/mfs_definitions.h"

/* Types definitions */

typedef enum { DCACHE_RESULT_OK,     /* Operation successful */
               DCACHE_RESULT_ERROR_NO_FILE, /* Can't find specified pathname */
               DCACHE_RESULT_ERROR_EXISTING, /* Can't create existing entry */
               DCACHE_RESULT_CANT_ADD, /* Memory error, etc */
               DCACHE_RESULT_ERROR_FATAL /* Cache corrupt */
             } dcache_result_t;

typedef void *dcache_key_t;
typedef void *dcache_browse_t;

typedef struct {
   mfs_import_object_locator_t locator;

   mfs_mode_t mode;
   mfs_uid_t uid;
   mfs_gid_t gid;
   mfs_index_object_size_t size;
   mfs_time_t mtime;

   char migrated_on[MAX_MIGRATED_ON_PATHNAME_LENGTH];
} dcache_stat_t;

/* Functional definitions */

/* NAME
 *    dcache_make
 * DESCRIPTION
 *    Creates a new DCACHE, removing if existing.
 * RESULT
 *    DCACHE_RESULT_OK or DCACHE_RESULT_ERROR_FATAL
 * NOTES
 * BUGS
 * TODO
 */
dcache_result_t dcache_make(char *pathname);

/* NAME
 *    dcache_init
 * DESCRIPTION
 *    Reads the DCACHE directory cache into memory, complaining if it is
 *    corrupted or inexistant.
 * RESULT
 *    NULL if cache corrupt or inexistant.
 * NOTES
 *    - Does not implement any locking.
 *    - This supposes we have enough room in memory for the directory cache.
 *    - mmap() is a possible implementation with no memory problem.
 * BUGS
 * TODO
 */
dcache_key_t dcache_init(char *pathname);

/* NAME
 *    dcache_deinit
 * DESCRIPTION
 *    Frees the memory directory cache after dcache_sync()ing. This
 *    means that key is no longer usable until the next dcache_init()
 *    which will return a new key.
 * RESULT
 *    dcache_result_t: DCACHE_RESULT_OK,
 *                     DCACHE_RESULT_ERROR_FATAL.
 * NOTES
 *    - Does not implement any locking.
 * BUGS
 * TODO
 */
dcache_result_t dcache_deinit(dcache_key_t key);

/* NAME
 *    dcache_stat
 * DESCRIPTION
 *    Gives information (in a sense similar to statf) about a
 *    migrated file (pathname), if it exists.
 * RESULT
 *    dcache_result_t: DCACHE_RESULT_OK
 *                     DCACHE_RESULT_ERROR_NO_FILE
 *                     DCACHE_RESULT_ERROR_FATAL
 * NOTES
 *    - Does not implement any locking.
 * BUGS
 * TODO
 */
dcache_result_t dcache_stat(dcache_key_t key,
                            const char *pathname,
                            dcache_stat_t *buf);

/* NAME
 *    dcache_create
 * DESCRIPTION
 *    Creates a new entry in the directory cache.
 * RESULT
 *    dcache_result_t: DCACHE_RESULT_OK
 *                     DCACHE_RESULT_ERROR_EXISTING
 *                     DCACHE_RESULT_ERROR_FATAL
 * NOTES
 *    - Does not implement any locking.
 * BUGS
 * TODO
 */
dcache_result_t dcache_create(dcache_key_t key,
                              const char *pathname,
                              dcache_stat_t buf);

/* NAME
 *    dcache_remove
 * DESCRIPTION
 *    Removes an entry from the directory cache.
 * RESULT
 *    dcache_result_t: DCACHE_RESULT_OK
 *                     DCACHE_RESULT_ERROR_NO_FILE
 *                     DCACHE_RESULT_ERROR_FATAL
 * NOTES
 *    - Does not implement any locking.
 * BUGS
 * TODO
 */
dcache_result_t dcache_remove(dcache_key_t key,
                              const char *pathname);

/* NAME
 *    dcache_sync
 * DESCRIPTION
 *    Ensures that the directory cache stored on disk is uptodate
 *    with the one in memory. All data must be on permanent storage
 *    on completion of this function.
 * RESULT
 *    dcache_result_t: DCACHE_RESULT_OK,
 *                     DCACHE_RESULT_ERROR_FATAL.
 * NOTES
 *    - Does not implement any locking.
 * BUGS
 *    - The ``all data on permanent storage'' is not implemented
 *      in the strict, fsync, sense.
 * TODO
 */
dcache_result_t dcache_sync(dcache_key_t key);

/* NAME
 *    dcache_browse
 * DESCRIPTION
 *    This function allows to look at the cache content in a
 *    hierarchic way. The first time it must be called with
 *    first == TRUE.
 * RESULT
 *    dcache_result_t: DCACHE_RESULT_OK
 *                     DCACHE_RESULT_ERROR_NO_FILE (at end)
 * NOTES
 *    - This function is intended for use by a ``directory-cache'' file
 *      system implementing soft-links to mounted tape filesystems or
 *      triggers for mounting them.
 * BUGS
 * TODO
 */
dcache_result_t dcache_browse(dcache_key_t key,
                              BOOL first,
                              dcache_browse_t *browsing_key,
                              dcache_stat_t *stat_buf, /* filled */
                              char *pathname,
                              int path_len);
                              
#endif


