/* Project mfs --- Migration file system
 * index.c -- Index management routines.
 * Creation: 19/12/1999
 * Update:   19/12/1999
 * V1.0 PV001
 * AUTHOR
 *    Marc SCHAEFER <schaefer@alphanet.ch>
 * LICENSE
 *    GPL
 * DESCRIPTION
 *    This implements the index management functions.
 *    There is only one index management per link (compile object).
 *    This interfaces with the dcache module.
 * NOTES
 *    - This is your typical middleware.
 * BUGS
 * TODO
 * BASED-ON
 *    Original work.
 * MODIFICATION-HISTORY
 *    19/12/99  schaefer  Created this file.
 * $Id: index.c,v 1.4 1999/12/30 10:04:16 schaefer Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include "../mfs/mfs_definitions.h"
#include "../mfs/mfs_interface.h"
#include "index_interface.h"
#define VFS_TYPES_H /* cheating */
#include "dcache/dcache_interface.h"

/* Private function definitions */

/* NAME
 *    index_internal_init
 * DESCRIPTION
 *    Internal initialization function
 * RESULT
 *    TRUE if database could be initialized (open).
 * NOTES
 * BUGS
 * TODO
 */
static BOOL index_internal_init(const char *name);

/* Private variables */

static dcache_key_t key = NULL; /* one and only
                                 * BUGS
                                 *    - Assuming things about key, see also
                                 *      the if below.
                                 */

BOOL index_init(const char *name) {
   if (key) {
      fprintf(stderr, "%s: already a key.\n", name);
      return FALSE;
   }
   else {
      if (index_internal_init(name)) {
         return TRUE;
      }
      else {
         dcache_result_t result;

         fprintf(stderr, "%s: creating.\n", name);
         result = dcache_make((char *) name);
         if (result == DCACHE_RESULT_OK) {
            return index_internal_init(name);
         }
         else {
            fprintf(stderr, "%s: can't create [0x%x].\n", name, result);
            return FALSE;
         }

         /* NOT REACHED */
      }
      /* NOT REACHED */
   }
   /* NOT REACHED */
}

BOOL index_deinit(void) {
   if (key) {
      dcache_result_t result;

      result = dcache_sync(key);
      if (result == DCACHE_RESULT_OK) {
         result = dcache_deinit(key);
	 if (result == DCACHE_RESULT_OK) {
	    key = NULL;
            return TRUE;
	 }
	 else {
	    fprintf(stderr, "can't deinit [0x%x].\n", result);
	    return FALSE;
	 }
         /* NOT REACHED */
      }
      else {
         fprintf(stderr, "can't sync [0x%x].\n", result);
         return FALSE;
      }
      /* NOT REACHED */
   }
   else {
      fprintf(stderr, "No index_init().\n");
      return FALSE;
   }
   /* NOT REACHED */
}

BOOL index_add(const char *name, mfs_import_data_t *data) {
   if (key) {
      dcache_result_t result;
      dcache_stat_t buf;

      buf.locator = data->locator;

      buf.mode = data->mode;
      buf.uid = data->uid;
      buf.gid = data->gid;
      buf.size = data->size;
      buf.mtime = data->mtime;

      memcpy(buf.migrated_on, data->migrated_on, sizeof(buf.migrated_on));

      result = dcache_create(key, name, buf);
      if (result == DCACHE_RESULT_OK) {
         return TRUE;
      }
      else {
         fprintf(stderr, "%s: can't create [0x%x].\n", name, result);
         return FALSE;
      }
      /* NOT REACHED */
   }
   else {
      fprintf(stderr, "No index_init().\n");
      return FALSE;
   }
   /* NOT REACHED */
}

BOOL index_get(const char *name, mfs_import_data_t *answer) {
   if (key) {
      dcache_result_t result;
      dcache_stat_t buf;

      result = dcache_stat(key, name, &buf);
      if (result == DCACHE_RESULT_OK) {
         answer->locator = buf.locator;

         answer->mode = buf.mode;
         answer->uid = buf.uid;
         answer->gid = buf.gid;
         answer->size = buf.size;
         answer->mtime = buf.mtime;

         memcpy(answer->migrated_on,
                buf.migrated_on,
                sizeof(answer->migrated_on));

         return TRUE;
      }
      else {
         fprintf(stderr, "%s: can't stat [0x%x].\n", name, result);
         return FALSE;
      }
      /* NOT REACHED */
   }
   else {
      fprintf(stderr, "No index_init().\n");
      return FALSE;
   }
   /* NOT REACHED */
}

BOOL index_remove(const char *name) {
   if (key) {
      dcache_result_t result;

      result = dcache_remove(key, name);
      if (result == DCACHE_RESULT_OK) {
         return TRUE;
      }
      else {
         fprintf(stderr, "%s: can't remove [0x%x].\n", name, result);
         return FALSE;
      }
      /* NOT REACHED */
   }
   else {
      fprintf(stderr, "No index_init().\n");
      return FALSE;
   }
   /* NOT REACHED */
}

/* Private function implementation */

static BOOL index_internal_init(const char *name) {
   key = dcache_init((char *) name);
   if (key) {
      return TRUE;
   }
   else {
      return FALSE;
   }
   /* NOT REACHED */
}


