/* Project mfs --- Migration file system
 * main.c -- mfsd main module
 * Creation: 17/10/1999
 * Update:   05/03/2000
 * V1.0 PV007
 * AUTHOR
 *    Marc SCHAEFER <schaefer@alphanet.ch>
 * LICENSE
 *    GPL
 * DESCRIPTION
 *    This implements the command line parsing and initialization for the
 *    mfsd daemon. The mfsd daemon communicates with the mfs filesystem and
 *    implements the user-level method for initiating migration,
 *    unmigration and maintains an index for the migrated files.
 *    It also supports multiple threads for unmigrating data in parallel.
 * NOTES
 *    - This is a TEST version, not to be used in a release. We need
 *      to modularize / remove debugging special cases, etc, add comments,
 *      exit codes, etc.
 * BUGS
 *    - We do system() with unchecked parameters on two occasions:
 *         - chattr, to reset the migrated flag
 *           -> this could be done through a ioctl(), to mfs, in the vein of
 *              the +A
 *         - for extracting files. This should be done through exec so that
 *           the file name is a controller parameter.
 *         - for cp -a (migr/unmigr). Could be done inside this software.
 *    - Extraction should be done in parallel.
 *    - Migration should be done synchronously: first +A, copy to temp,
 *      truncate. Then parallel when temp starts to get beyond a certain
 *      threshold, through migrd to spool or tape.
 *    - Still a few test cases (SPECIAL_FILE_NAME, etc) buried in the
 *      code to be removed when everything mostly works.
 *    - We should really split the differing functionnality in different
 *      modules (e.g. threads should not be able to peak into main).
 *      But first check if we keep this threading structure or change it
 *      by exec() of other executables (ie changer entitites).
 * TODO
 *    - Split this file into separate entities.
 * BASED-ON
 *    Original work.
 * MODIFICATION-HISTORY
 *    17/10/99  schaefer  Created this file.
 *    11/02/00  schaefer  General cleanup.
 *    24/02/00  schaefer  Added conditional compiled-in select().
 *    26/02/00  schaefer  select() is no longer an option
 *    27/02/00  schaefer  Ugly workaround for two FIST bugs.
 *    27/02/00  schaefer  Using NOTIFY_LIVES ioctl() to request mfs info.
 *    05/03/00  schaefer  Started multithread implementation.
 * $Id: main.c,v 1.33 2000/02/27 18:12:15 schaefer Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <signal.h>
#include <dirent.h>
#include <errno.h>
#include "../mfs/mfs_definitions.h"
#include "../mfs/mfs_interface.h"
#include "index_interface.h"

#define MAX(x, y) (((x) > (y)) ? (x) : (y))

#define TEST_FILE "directory/special-file"
#define SPECIAL_STRING "The Quick Brown Fox Jumps Over The Lazy Dog.\n"

/* Defines the place where a file is migrated when it's only on temporary
 * storage.
 */
#define MFS_SPECIAL_TEMP_MIGRATED_ON "--TEMP--"

/* Type definitions */

/* NOTES
 *    - The idea behind all this is to allow a certain number of
 *      requests to be queued, either on a WAITING queue, a
 *      CONFLICT, or a RUNNING queue. The WAITING queue is used
 *      for requests which couldn't be assigned to a child process
 *      because all are busy. The CONFLICT queue is for operation
 *      that conflict with currently RUNNING or WAITING requests.
 *      The RUNNING queue is for requests currently being run by
 *      a child. There is also a FREE queue for not used slots
 *      (so we don't constantly malloc()/free(). Current implementation
 *      is a simple bounded table (however dynamically allocated at boot
 *      time): kernel also has a boundary on requests anyway.
 * TODO
 *    - Define protocol more closely, especially for deadlocks.
 */

typedef enum {
   MFSD_CHILD_STATE_NOT_LAUNCHED, /* No child connected */
   MFSD_CHILD_STATE_INITIALIZING,
   MFSD_CHILD_STATE_WAITING,
   MFSD_CHILD_STATE_RUNNING
} mfsd_child_state_t;

/* Indexes */
#define MFSD_OBJECT_NUMBER_NONE 255
typedef unsigned char mfsd_object_number_t;
typedef unsigned char mfsd_child_number_t;

typedef struct {
   pid_t pid;

   fd_t to;
   fd_t from;

   mfsd_child_state_t state;

   mfsd_object_number_t object_number; /* MFSD_OBJECT_NUMBER_NONE for not
                                        * assigned.
                                        */
} mfsd_child_t;

typedef enum {
   MFSD_OPERATION_NONE,  /* free slot */
   MFSD_OPERATION_UNMIGRATE
} mfsd_operation_type_t;

typedef struct {
   mfsd_operation_type_t operation;
   char path_name[MAX_PATHNAME_LENGTH];
   mfs_import_data_t import_data;
} mfsd_operation_t;

typedef struct {
   mfsd_operation_t operation;

   mfsd_child_t *handler; /* NULL means: not yet assigned
                          * (if !MFSD_OPERATION_NONE).
                          */
} mfsd_object_t;

typedef struct {
   mfsd_operation_t operation;
} mfsd_pipe_request_t;

typedef enum {
   MFSD_PIPE_ANSWER_RESULT_GOOD,
   MFSD_PIPE_ANSWER_RESULT_ERROR
} mfsd_pipe_answer_result_t;

typedef enum {
   MFSD_PIPE_OPERATION_ADD_INDEX,
   MFSD_PIPE_OPERATION_REMOVE_INDEX
} mfsd_pipe_operation_t;

typedef struct {
   mfsd_pipe_answer_result_t result;
   mfsd_pipe_operation_t operation;
   /* @@@ */
} mfsd_pipe_answer_t;

/* When mass-migrating, the daemon first registers all the files he is
 * going to migrate to mfsd. Then it mass migrates. Then at the end it
 * tells mfsd it's finished, with the list of non succeeded files.
 * mfsd, during the migrate, removes from the list of files any file
 * modified or unlinked, and also removes from that list the non successful
 * ones from migrd. Atomically, the updated list contains the correctly
 * migrated files and thus the index can be updated, and the files in
 * temp are candidate for deletion when temp becomes full. Maybe a bit
 * in index:
 *    VALUE (or e.g. name --TEMP--)  BIT    DESCRIPTION
 *    0                              0|1    INVALID
 *    1                              0|1    file is on temp storage
 *    2                              0      file migrated on tape or other
 *                                   1      and additionnaly still in temp
 * We still need to define more the interactions (e.g. how migrd
 * communicates with mfsd, since it can be launched by mfsd or cron, etc).
 * @@@
 */

typedef struct {
   char *program_name;

   char *underlying_fs;
   char *mfs_mount_point;
   char *temporary_directory;
   char *mfs_index_name;

   BOOL do_finish; /* modified by signal
                    * BUGS
                    *    - in theory would need volatile.
                    */

   fd_t f; /* for mfs special ioctl()s */

   /* Minimum migration to do when low on space */
   mfs_index_object_size_t minimum_batch_threshold;

   /* Don't migrate files smaller than this size in any case */
   mfs_index_object_size_t minimum_size_migrate;

   mfsd_child_number_t children_count;
   mfsd_child_t *children;
} mfsd_config_t;

/* Private function definitions */

/* NAME
 *    do_processing
 * DESCRIPTION
 *    Gets events from mfs, process them.
 * RESULT
 *    standard exit code
 * NOTES
 * BUGS
 * TODO
 */
static int do_processing(mfsd_config_t *mfsd_config);

/* NAME
 *    do_unmigrate
 * DESCRIPTION
 *    Unmigrate a file from off-line storage.
 * RESULT
 *    TRUE if unmigration succcessful.
 * NOTES
 *    - file_name is relative to the mfs filesystem.
 * BUGS
 * TODO
 */
static BOOL do_unmigrate(const mfsd_config_t *mfsd_config,
                         const char *path_name,
                         mfs_import_object_locator_t locator,
                         const char *migrated_on);

/* NAME
 *    set_unset_migrated
 * DESCRIPTION
 *    Sets or resets the migration state of a file,
 *    depending on the boolean set.
 * RESULT
 *    BOOL TRUE if operation succeeded.
 * NOTES
 *    - file name is relative to underlying fs.
 * BUGS
 *    - Currently implemented with system() and chattr since
 *      couldn't find any system call. I found the ioctl(),
 *      but now I am too lazy. Should really be implemented by
 *      a generic ioctl to mfs like it's done for atomic migration.
 * TODO
 */
static BOOL set_unset_migrated(const mfsd_config_t *mfsd_config,
                               const char *file,
                               BOOL set);

/* NAME
 *    fake_migration
 * DESCRIPTION
 *    Does a migration for testing purposes only.
 * RESULT
 *    BOOL TRUE if operation succeeded.
 * NOTES
 *    - file name is relative to underlying fs.
 * BUGS
 *    - This is also used, at this time, by
 *      MFS_COMMAND_CREATE_MIGRATED_OBJECT, and this means
 *      we have a window where something bad could happen.
 *      See comments in MFS_COMMAND_CREATE_MIGRATED_OBJECT.
 * TODO
 */

static BOOL fake_migration(mfsd_config_t *mfsd_config, const char *file);

/* NAME
 *    create_file
 * DESCRIPTION
 *    creat(2) wrapper, in the underlying fs.
 * RESULT
 *    BOOL TRUE if operation succeeded.
 * NOTES
 *    - file_name is relative to the underlying filesystem.
 * BUGS
 * TODO
 */
static BOOL create_file(mfsd_config_t *mfsd_config, const char *file);

/* NAME
 *    signal_handler
 * DESCRIPTION
 *    Signal handler for SIGINT and SIGTERM, so that termination
 *    is done properly, including saving the index file. Sets
 *    the global do_finish in the mfsd_config variable.
 * RESULT
 *    NONE
 * NOTES
 * BUGS
 * TODO
 */
static void signal_handler(int value);

/* NAME
 *    has_extension
 * DESCRIPTION
 *    Returns TRUE if the specified file name exactly ends with the
 *    specified extension.
 * RESULT
 *    TRUE if matches.
 * NOTES
 * BUGS
 * TODO
 */
static BOOL has_extension(const char *file_name, const char *extension);

/* NAME
 *    do_mass_migration
 * DESCRIPTION
 *    Migrates files in the mfs filesystem until MINIMUM_BATCH_THRESHOLD
 *    bytes are freed. Only migrate not yet migrated files which
 *    are not open (this is atomically tested) and which are bigger
 *    that MINIMUM_SIZE_MIGRATE.
 * RESULT
 *    TRUE if some bytes were migrated
 * NOTES
 *    - The non LOW_SPACE version of this should be done in mfs_migrd
 *      from time to time if rules on disk space are matched.
 * BUGS
 *    - Presumably MINIMUM_BATCH_THRESHOLD should be smaller than
 *      the maximum type representable by mfs_index_object_size_t, and
 *      bigger than zero.
 * TODO
 */
static BOOL do_mass_migration(mfsd_config_t *mfsd_config);

/* NAME
 *    migration_candidate
 * DESCRIPTION
 *    Return TRUE if the file is a migration candidate, ie it is
 *    bigger than MINIMUM_SIZE_MIGRATE.
 * RESULT
 *    TRUE if a migration candidate
 * NOTES
 *    - Does not do the atomic migration/open file test, which must
 *      be done atomically at the time the decision was already made.
 *    - file_name is absolute in the underlying filesystem.
 * BUGS
 * TODO
 *    - In the future, should be extended to use other criteria for
 *      migration, such as last data access time, etc. We probably
 *      have then to return a priority value and the caller should
 *      create an ordered list of (file, priority), and select the
 *      first(s).
 */
static BOOL migration_candidate(mfsd_config_t *mfsd_config,
                                const char *file_name,
                                const struct stat *stat_buf);

/* NAME
 *    migration_do
 * DESCRIPTION
 *    Does the actual migration. If it fails, the file is put back in
 *    a stable state.
 * RESULT
 *    TRUE if migration succeeded
 * NOTES
 *    - file_name is relative to the underlying filesystem.
 * BUGS
 * TODO
 */
static BOOL migration_do(mfsd_config_t *mfsd_config, const char *file_name);

/* NAME
 *    get_file_size
 * DESCRIPTION
 *    Gets the named file size.
 * RESULT
 *    TRUE if the file size was correctly set.
 * NOTES
 *    - file_name is relative to the underlying filesystem.
 * BUGS
 * TODO
 */
static BOOL get_file_size(mfsd_config_t *mfsd_config,
                          const char *file_name,
                          mfs_index_object_size_t *size);

/* NAME
 *    remove_file_from_temp
 * DESCRIPTION
 *    Removes a file in the temporary directory.
 * RESULT
 *    TRUE if the file was correctly removed.
 * NOTES
 *    - file_name is relative to the underlying filesystem.
 * BUGS
 *    - done with system(). Well, this is a hack anyway.
 * TODO
 */
static BOOL remove_file_from_temp(const mfsd_config_t *mfsd_config,
                                  const char *file_name);

/* NAME
 *    copy_file_to_temp
 * DESCRIPTION
 *    Copies a file being migrated to the temporary directory. This may
 *    involve starting a synchronous off-line write if the temp directory
 *    starts to get full. May fail if there is a serious unrecoverable
 *    error (such as off-line storage not available if the temporary
 *    directory becomes full). May also start an asynchronous off-line
 *    write if the temporary directory usage is above a certain threshold.
 * RESULT
 *    TRUE if the file was correctly copied to the temporary directory.
 * NOTES
 *    - file_name is relative to the underlying filesystem.
 * BUGS
 * TODO
 */
static BOOL copy_file_to_temp(mfsd_config_t *mfsd_config,
                              const char *file_name);

/* NAME
 *    truncate_file
 * DESCRIPTION
 *    Truncates the file name in a correct way.
 * RESULT
 *    TRUE if the file could be truncated correctly. FALSE if the
 *    files was not truncated.
 * NOTES
 *    - file_name is relative to the mfs filesystem.
 * BUGS
 * TODO
 */
static BOOL truncate_file(mfsd_config_t *mfsd_config,
                          const char *file_name);

/* NAME
 *    remount_rw_workaround
 * DESCRIPTION
 *    remount the specified filesystem read-write, this presumably is
 *    used to work-around the capacity used FIST bug after rm, and
 *    the truncate FIST bug.
 * RESULT
 *    TRUE if the fs could be remounted.
 * NOTES
 *    - fs is absolute and must be a filesystem in /etc/fstab
 * BUGS
 * TODO
 */
static BOOL remount_rw_workaround(mfsd_config_t *mfsd_config,
                                  const char *fs,
                                  BOOL please_fork);

/* NAME
 *    traverse_and_migrate
 * DESCRIPTION
 *    Traverse a complete directory, possibly migrating the files
 *    upto the certain total of bytes if they comply on the constraints,
 *    and recursively calling itself for directories.
 * RESULT
 *    Number of migrated bytes.
 * NOTES
 *    - This is recursive. Every recursion will take a few K of
 *      stack. However, the recursion is statically limited by the
 *      path name size.
 *    - current_directory_path is absolute in the underlying fs.
 * BUGS
 * TODO
 */
static mfs_index_object_size_t
   traverse_and_migrate(mfsd_config_t *mfsd_config,
                        const char *current_directory_path,
                        mfs_index_object_size_t migrated_bytes);

/* NAME
 *    create_directory_path
 * DESCRIPTION
 *    Does the equivalent of a ``mkdir -p'' in a faster/less dangerous
 *    way.
 * RESULT
 *    TRUE if succeeded
 * NOTES
 *    - name is absolute
 * BUGS
 *    - This comes from ../mfs_import and should really be put
 *      in a generic library.
 * TODO
 */
static BOOL create_directory_path(const char *name);

/* NAME
 *    simplify_path
 * DESCRIPTION
 *    Simplify the given path by
 *       - removing any embedded ^./ (^ denotes starts of string)
 *         or ^../
 *       - replacing /./ by /
 *       - replacing any // by /
 *       - replacing /../ by going up, upto wherever possible.
 *       - removing /mfs-mount-point/ prefix.
 * RESULT
 *    NONE
 * NOTES
 *    - The passed string will be modified.
 * BUGS
 * TODO
 *    - Evaluate if we need to fail, e.g. when too many ../
 */
static void simplify_path(char *path, const char *mfs_mount_point);

/* NAME
 *    shift_string
 * DESCRIPTION
 *    Removes the first count character of string by safely copying.
 * RESULT
 *    NONE
 * NOTES
 * BUGS
 * TODO
 */
static void shift_string(char *string, int count);

/* NAME
 *    string_scan
 * DESCRIPTION
 *    Returns the pointer in string of the first occurence of part, NULL
 *    if none.
 * RESULT
 *    char * or NULL if no match.
 * NOTES
 * BUGS
 * TODO
 */
static char *string_scan(const char *string, const char *part);

/* NAME
 *    queue_request
 * DESCRIPTION
 *    Queues a request either to one of the children, or to
 *    a global queue.
 * RESULT
 *    BOOL
 * NOTES
 * BUGS
 * TODO
 */
static BOOL queue_request(const mfsd_config_t *mfsd_config,
                          const char *path_name,
                          mfs_import_data_t *import_data);

/* NAME
 *    queue_allocate_object
 * DESCRIPTION
 *    Allocates a queue object.
 * RESULT
 *    mfsd_object_t *object, or NULL.
 * NOTES
 * BUGS
 * TODO
 */
mfsd_object_t *queue_allocate_object(const mfsd_config_t *mfsd_config);

/* NAME
 *    send_request
 * DESCRIPTION
 *    Sends a request to the specified child.
 * RESULT
 *    BOOL, TRUE if sent.
 * NOTES
 * BUGS
 * TODO
 */
BOOL send_request(const mfsd_config_t *mfsd_config,
                  mfsd_child_t *child,
                  mfsd_object_t *object);

/* NAME
 *    internal_queue
 * DESCRIPTION
 *    Queues an object.
 * RESULT
 *    NONE
 * NOTES
 * BUGS
 * TODO
 */
void internal_queue(const mfsd_config_t *mfsd_config,
                    mfsd_object_t *object);

/* NAME
 *    threads_launch
 * DESCRIPTION
 *    Based on the size of the kernel fs driver queues, creates
 *    mfsd child processes (mfsd threads, in mfs terminology),
 *    and creates the appropriate queues in mfsd_config.
 * RESULT
 *    TRUE if all threads launched correctly, else cleaned up already.
 * NOTES
 * BUGS
 */
static BOOL threads_launch(mfsd_config_t *mfsd_config, mfs_info_t mfs_info);

/* NAME
 *    threads_fd_set
 * DESCRIPTION
 *    Sets the corresponding file descriptor set bits corresponding to
 *    threads that are active.
 * RESULT
 *    fd_t, the last set fd.
 * NOTES
 * BUGS
 */
static fd_t threads_fd_set(mfsd_config_t *mfsd_config, fd_set *fds);

/* NAME
 *    threads_fd_do
 * DESCRIPTION
 *    Does the appropriate thread handling on the ready-to-serve
 *    threads (represented by the fd_set).
 * RESULT
 *    NONE
 * NOTES
 * BUGS
 */
static void threads_fd_do(mfsd_config_t *mfsd_config, fd_set *fds);

/* NAME
 *    threads_stop
 * DESCRIPTION
 *    Stops all not yet finished threads cleanly.
 * RESULT
 *    NONE
 * NOTES
 * BUGS
 *    - This is probably better in the main loop, no longer accepting
 *      kernel queries, but finishing properly and accepting children
 *      queries. Would then return TRUE when no threads remaining.
 */
static void threads_stop(mfsd_config_t *mfsd_config);

/* NAME
 *    thread_launch
 * DESCRIPTION
 *    Launches a thread.
 * RESULT
 *    BOOL
 * NOTES
 * BUGS
 * TODO
 */
static BOOL thread_launch(mfsd_child_t *child);

/* NAME
 *    thread_stop
 * DESCRIPTION
 *    Stops a thread.
 * RESULT
 *    NONE
 * NOTES
 * BUGS
 * TODO
 */
static void thread_stop(mfsd_child_t *child);

/* NAME
 *    thread_initialize
 * DESCRIPTION
 *    Initializes the thread protocol
 * RESULT
 *    TRUE if succeeds. If false, the thread is presumably dead.
 * NOTES
 * BUGS
 * TODO
 */
static BOOL thread_initialize(mfsd_child_t *child);

/* NAME
 *    mfsd_thread
 * DESCRIPTION
 *    Work-horse of the thread.
 * RESULT
 *    NONE
 * NOTES
 * BUGS
 * TODO
 */
static void mfsd_thread(const mfsd_config_t *mfsd_config,
                        fd_t commands,
                        fd_t answers);

/* NAME
 *    thread_get_request
 * DESCRIPTION
 *    Receive a request from main.
 * RESULT
 *    BOOL, TRUE if no error.
 * NOTES
 * BUGS
 * TODO
 */
static BOOL thread_get_request(fd_t in, mfsd_operation_t *operation);

/* NAME
 *    thread_send_result
 * DESCRIPTION
 *    Send a result to main.
 * RESULT
 *    BOOL, TRUE if no error.
 * NOTES
 * BUGS
 * TODO
 */
static BOOL thread_send_result(fd_t out, mfsd_pipe_answer_result_t *result);

/* Private variables */

static mfsd_config_t mfsd_config; /* here because of signal trick */

/* Function implementation */

int main(int argc, char **argv) {
   int result;

   if (argc == 5) {
      mfsd_config.program_name = argv[0];
      mfsd_config.underlying_fs = argv[1];
      mfsd_config.mfs_mount_point = argv[2];
      mfsd_config.temporary_directory = argv[3];
      mfsd_config.mfs_index_name = argv[4];
      mfsd_config.minimum_batch_threshold = MINIMUM_BATCH_THRESHOLD;
      mfsd_config.minimum_size_migrate = MINIMUM_SIZE_MIGRATE;

      printf("%s: underlying fs %s; mfs %s; temp %s; index %s; "
             "minimum low-space batch migration %ld; "
             "minimum file size to migrate %ld.\n",
             mfsd_config.program_name,
             mfsd_config.underlying_fs,
             mfsd_config.mfs_mount_point,
             mfsd_config.temporary_directory,
             mfsd_config.mfs_index_name,
             mfsd_config.minimum_batch_threshold,
             mfsd_config.minimum_size_migrate);

      if (signal(SIGINT, signal_handler) != SIG_ERR) {
         if (signal(SIGTERM, signal_handler) != SIG_ERR) {
	    result = do_processing(&mfsd_config);
            /* We leave the signal handlers as they are */
         }
         else {
            fprintf(stderr, "%s: error setting SIGTERM handler.\n", argv[0]);
            result = 1;
         }
      }
      else {
         fprintf(stderr, "%s: error setting SIGINT handler.\n", argv[0]);
         result = 1;
      }
   }
   else {
      fprintf(stderr, "%s underlying-fs mfs-mountpoint temp-directory index\n",
              argv[0]);
      fprintf(stderr, "%s: bad args.\n", argv[0]);
      result = 2;
   }

   return result;
}

/* Private function implementation */

static void signal_handler(int value) {
   mfsd_config.do_finish = TRUE;
}

static int do_processing(mfsd_config_t *mfsd_config) {
   mfs_request_t request;
   mfs_answer_t answer;
   int err = 0;
   BOOL finished = FALSE;
   char buffer[MAX_PATHNAME_LENGTH];

   if (index_init(mfsd_config->mfs_index_name)) {
      if (snprintf(buffer,
		   sizeof(buffer),
		   "%s/%s",
		   mfsd_config->mfs_mount_point,
		   MFS_SPECIAL_MFSD_FILE)
	  > -1) {
         mfs_info_t mfs_info;

         mfs_info.protocol_version = MFS_PROTOCOL_VERSION;

	 mfsd_config->f = open(buffer, O_RDWR, 0600);
         /* NOTES
          *    - When using select() before the GET_REQUEST ioctl(),
          *      the file was not yet registered with mfs. This is
          *      anyway now required by the new interface.
          */
         if (ioctl(mfsd_config->f,
                   MFS_IOCTL_NOTIFY_LIVES,
                   &mfs_info)) {
            perror("ioctl(MFS_IOCTL_NOTIFY_LIVES)");
            close(mfsd_config->f); /* result ignored */
            err = 1;
         }
         else if (mfsd_config->f != -1 ) {
            printf("%s: queue depth %d; reserved %d; protocol %d.\n",
                   mfsd_config->program_name,
                   mfs_info.queue_depth,
                   mfs_info.reserved_entries,
                   mfs_info.protocol_version);

            printf("%s: launching threads.\n",
                   mfsd_config->program_name);

            if (!threads_launch(mfsd_config, mfs_info)) {
               fprintf(stderr, "%s: error launching threads.\n",
                               mfsd_config->program_name);
               err = 1;
            }
            /* Setup the test environment */
	    else if (fake_migration(mfsd_config, TEST_FILE)) {

               /* Main processing loop */

	       while (!err && !finished && !mfsd_config->do_finish) {
                  mfs_import_data_t temp;
                  /* NOTES
                   *    - Commands will be executed either in parallel,
                   *      by threads, or by us.
                   */
                  BOOL command_in_multitask = FALSE;

		  memset((char *) &request, 0, sizeof(request));
		  memset((char *) &answer, 0, sizeof(answer));

		  /* It is the first ioctl() which creates the association.
		   * After this, if the file is closed, mfsd is taken for
		   * dead by mfs and errors are returned.
		   */

                  /* NOTES
                   *    - This is the first step towards mfsd multithreading.
                   */
                  fprintf(stderr, "mfsd: calling select().\n");
                  fflush(stderr);
                  while (TRUE) {
                     fd_set exceptfds;
                     fd_set readfds;
                     int retval;
                     fd_t max_fd = 0;

                     FD_ZERO(&exceptfds);
                     FD_SET(mfsd_config->f, &exceptfds);

                     FD_ZERO(&readfds);
                     max_fd = MAX(threads_fd_set(mfsd_config, &readfds),
                                  mfsd_config->f);

                     /* Wait indefinitely for event */
                     retval = select(max_fd + 1,
                                     NULL,
                                     NULL,
                                     &exceptfds,
                                     NULL);
                     if (retval > 0) {
                        /* Do threads processing, if any */
                        threads_fd_do(mfsd_config, &readfds);

                        if (FD_ISSET(mfsd_config ->f, &exceptfds)) {
                           break; /* out of while */
                        }
                     }
                     else if (retval == -1) {
                        perror("select()");
                        finished = TRUE;
                        break; /* out of while */
                     }
                     else {
                        fprintf(stderr, "mfsd: strange: select() says 0.\n");
                        fflush(stderr);
                     }                     
                  }
                  fprintf(stderr, "mfsd: return from select().\n");
                  fflush(stderr);

                  if (finished) {
 		     break; /* out of while */
                  }

		  if (ioctl(mfsd_config->f,
                            MFS_IOCTL_GET_REQUEST_BLOCK,
                            &request)) {
	             perror("ioctl(GET)");
		     err = 1;
		  }

		  if (err) {
		     break;
		  }

		  printf("got request 0x%x, private 0x%lx, pathname=%s.\n",
			 request.command,
			 (unsigned long int) request.private,
			 request.path_name);

                  simplify_path(request.path_name,
                                mfsd_config->mfs_mount_point);

                  printf("simplified path name: %s.\n", request.path_name);

		  switch (request.command) {
		     case MFS_COMMAND_GET_INDEX_OBJECT:
			/* Temporary test implementation */
			if (!strcmp(request.path_name, TEST_FILE)) {
			   answer.size = sizeof(SPECIAL_STRING);

			   printf("GET: positive answer (TEST CASE).\n");
			   answer.status = MFS_STATUS_OK;
			}
                        else if (index_get(request.path_name, &temp)) {
			   printf("GET: positive answer (FROM INDEX).\n");
                           answer.size = temp.size;
			   answer.status = MFS_STATUS_OK;
                        }
			else {
			   printf("GET: negative answer.\n");
			   answer.status = MFS_STATUS_OBJECT_NOT_FOUND;
			}
			break;
		     case MFS_COMMAND_SYSTEM_LOW_IN_SPACE:
			fprintf(stderr, "%s: system low in space.\n",
                                mfsd_config->program_name);

                        if (do_mass_migration(mfsd_config)) {
                           answer.status = MFS_STATUS_OK;
                        }
                        else {
			   answer.status = MFS_STATUS_NO_FREED_SPACE;
                        }
			break;
		     case MFS_COMMAND_UNMIGRATE_OBJECT:
			/* Temporary test implementation */
			if (!strcmp(request.path_name, TEST_FILE)) {
			   printf("UNMIGRATE: creating data (TEST CASE).\n");

			   if (snprintf(buffer,
					sizeof(buffer),
					"%s/%s",
					mfsd_config->underlying_fs,
					TEST_FILE) > -1) {
			      int f2;

			      f2 = open(buffer, O_WRONLY, 0666);
			      if (f2 != -1) {
				 if (write(f2,
					   SPECIAL_STRING,
					   sizeof(SPECIAL_STRING))
				     == sizeof(SPECIAL_STRING)) {
				    if (close(f2)) {
				       perror("close()");
				       err = 1;
				       answer.status = MFS_STATUS_IO_ERROR;
				    }
				    else {
				       if (set_unset_migrated(mfsd_config,
                                                              TEST_FILE,
                                                              FALSE)) {
					  answer.status = MFS_STATUS_OK;
				       }
				       else {
					  fprintf(stderr,
						  "%s: can't reset migred.\n",
						  mfsd_config->program_name);
					  err = 1;
					  answer.status = MFS_STATUS_IO_ERROR;
				       }
				    }
				 }
				 else {
				    perror("write()");
				    err = 1;
				    answer.status = MFS_STATUS_IO_ERROR;
				 }
			      }
			      else {
				 perror("open()");
				 err = 1;
				 answer.status = MFS_STATUS_IO_ERROR;
			      }
			   }
			   else {
			      fprintf(stderr, "path too long %s.\n", buffer);
			   }	
			}
                        else if (index_get(request.path_name, &temp)) {
			   printf("UNMIGRATE: positive (FROM INDEX).\n");
                           printf("file %s size %ld locator %ld mo %s.\n",
                                  request.path_name,
                                  temp.size,
                                  temp.locator,
                                  temp.migrated_on);
                           /* This will also lock the object */
                           if (queue_request(mfsd_config,
                                             request.path_name,
                                             &temp)) {
                              command_in_multitask = TRUE;
                           }
                           else {
                              fprintf(stderr,
                                      "%s: can't queue request %s.\n",
                                      mfsd_config->program_name,
                                      request.path_name);
                              answer.status = MFS_STATUS_IO_ERROR;
                           }
                        }
			else {
			   printf("UNMIGRATE: negative answer.\n");
			   answer.status = MFS_STATUS_OBJECT_NOT_FOUND;
			}
			break;
		     case MFS_COMMAND_SHUTDOWN:
                        fprintf(stderr, "%s: SHUTDOWN.\n",
                                mfsd_config->program_name);
			finished = TRUE; /* do NOT answer */
			break;
		     case MFS_COMMAND_CREATE_MIGRATED_OBJECT:
			fprintf(stderr,
				"%s: object %s available on %s, size %ld.\n",
				mfsd_config->program_name,
				request.path_name,
				request.import.migrated_on,
				request.import.size);
			/* BUGS
			 *    - not using uid/gid/mode/mtime, etc.
			 *    - not discriminating error/exists.
                         *    - to avoid all races, the file should be
                         *      created migrated. We have to add a new
                         *      mfs ioctl: create-migrated-empty-file().
                         *      For now we use create_file()/fake_migration(),
                         *      but there is a window of vulnerability.
			 */

			if (index_add(request.path_name,
				      &request.import)) {
			   if (create_file(mfsd_config, request.path_name)) {
			      if (fake_migration(mfsd_config,
                                                 request.path_name)) {
				 answer.status = MFS_STATUS_OK;
			      }
			      else {
				 fprintf(stderr,
					 "%s: can't migrate %s.\n",
					 mfsd_config->program_name,
					 request.path_name);
				 answer.status = MFS_STATUS_IO_ERROR;
			      }
			   }
			   else {
			      fprintf(stderr,
				      "%s: can't create %s.\n",
				      mfsd_config->program_name,
				      request.path_name);
			      answer.status = MFS_STATUS_IO_ERROR;
			   }
			}
			else {
			   /* BUGS
			    *    - we need to discriminate between
			    *      no room, exists already, etc.
			    */
			   fprintf(stderr, "%s: index_add(%s) failed.\n",
				   mfsd_config->program_name,
                                   request.path_name);
			   answer.status = MFS_STATUS_IO_ERROR;
			}
			break;
  		     case MFS_COMMAND_DELETE_INDEX_OBJECT:
                        fprintf(stderr, "%s: INDEX REMOVE %s.\n",
                                mfsd_config->program_name,
                                request.path_name);
			if (!index_remove(request.path_name)) {
			   fprintf(stderr,
                                   "%s: MFS_COMMAND_DELETE_INDEX_OBJECT"
                                   "index_remove() failed.\n",
				   mfsd_config->program_name);
			}
                        else {
                           /* No need to truncate the file */
                           set_unset_migrated(mfsd_config,
                                              request.path_name,
                                              FALSE);
                           /* NOTES
                            *    - we do remove the file from the temporary
                            *      directory, since it could be partly
                            *      migrated (ie to temp only).
                            * ASSUMPTION
                            *    - noone else is migrating that file, else
                            *      risk of deleting a migrated file. This
                            *      should be ensured when multithreaded
                            *      mfs is implemented (including for
                            *      mass migration).
                            */
                           remove_file_from_temp(mfsd_config,
                                                 request.path_name); /* res.
                                                                      * ign.
                                                                      */
                        }
                        /* In any case, succeed, since parallelism doesn't
                         * guarantee there aren't many parallel requests
                         * doing the same.
                         */
			break;
                     case MFS_COMMAND_MIGRATE_OBJECT:
                        fprintf(stderr, "%s: MIGRATE OBJECT %s.\n",
                                mfsd_config->program_name,
                                request.path_name);
                        if (!migration_do(mfsd_config, request.path_name)) {
                           answer.status = MFS_STATUS_IO_ERROR;
                        }
                        break;
                     case MFS_COMMAND_WORK_AROUND:
                        fprintf(stderr, "%s: WORK AROUND.\n",
                                mfsd_config->program_name);
                        /* Do the two work-arounds, never returning any
                         * error. This presumably fix the non freed
                         * space bug.
                         */

                        if (!remount_rw_workaround(mfsd_config,
                                                   mfsd_config
                                                      ->mfs_mount_point,
                                                   TRUE)) {
                           fprintf(stderr, "workaround: mfs failed.\n");
                        }
                        break;
		     default:
			printf("unsupported cmd 0x%x.\n", request.command);
			answer.status = MFS_STATUS_OBJECT_NOT_FOUND;   
			break;
		  }

		  if (!finished) {
                     if (!command_in_multitask) {
			answer.private = request.private;

			if (ioctl(mfsd_config->f,
				  MFS_IOCTL_SEND_ANSWER_BLOCK,
				  &answer)) {
			   perror("ioctl(SEND)");
			   err = 1;
			}
                     }
		  }
		  else {
		     fprintf(stderr,
			     "%s: shutting down by closing.\n",
			     mfsd_config->program_name);
		  }
	       }

               printf("%s: stopping all threads.\n",
                      mfsd_config->program_name);
	       threads_stop(mfsd_config);
	    }
	    else {
	       fprintf(stderr, "%s: fake_migration() failed.\n",
                       mfsd_config->program_name);
	       err = 1;
	    }
	    close(mfsd_config->f); /* result ignored */
	 }
	 else {
	    err = 1;
	    perror("open()");
	 }
      }
      else {
	 fprintf(stderr, "%s: path too long.\n", mfsd_config->program_name);
	 err = 1;
      }
      if (!index_deinit()) {
          fprintf(stderr, "%s: can't deinit index.\n",
                  mfsd_config->program_name);
          err = 1;
      }
      else {
         printf("%s: program terminated properly.\n",
	        mfsd_config->program_name);
      }
   }
   else {
      /* BUGS
       *    - Should tell why.
       *    - Index name should not be in stone.
       */
      fprintf(stderr, "%s: can't init index.\n", mfsd_config->program_name);
      err = 1;
   }

   return err;
}

/* Private function implementation */

static 
BOOL set_unset_migrated(const mfsd_config_t *mfsd_config,
                        const char *file,
                        BOOL set) {
   if (set) {
      mfs_atomic_migrate_t migrate;

      if (strlen(file) < sizeof(migrate.path_name)) {
         memcpy(migrate.path_name, file, sizeof(migrate.path_name));
         if (ioctl(mfsd_config->f,
                   MFS_IOCTL_ATOMIC_MIGRATE,
                   &migrate) == 0) {
            return TRUE;
         }
         else {
            perror("MFS_IOCTL_ATOMIC_MIGRATE");
            return FALSE;
         }
      }
      else {
	 fprintf(stderr, "%s: file name too long (%s).\n",
                 mfsd_config->program_name,
                 file);
	 return FALSE;
      }
   }
   else {
      char buffer[MAX_PATHNAME_LENGTH];

      /* NOTES
       *   - This implementation was the previous one (non atomic, no
       *     check for open files on SET). It's perfectly adequate for
       *     UNSET. We should one day clean it up, removing the SET case.
       */
      if (snprintf(buffer,
		   sizeof(buffer),
		   "chattr %cA %s/%s",
		   set ? '+' : '-',
		   mfsd_config->underlying_fs,
		   file)
	  > -1) {
	 return (system(buffer) == 0);
      }
      else {
	 fprintf(stderr, "%s: snprintf() failed.\n",
                 mfsd_config->program_name);
	 return FALSE;
      }
   }
}

static BOOL fake_migration(mfsd_config_t *mfsd_config, const char *file) {
   /* BUGS
    *    - See function definition.
    */

   /* Playing on && C order */
   return (truncate_file(mfsd_config, file)
           && set_unset_migrated(mfsd_config, file, TRUE));
}

static BOOL create_file(mfsd_config_t *mfsd_config, const char *file) {
   char buffer[MAX_PATHNAME_LENGTH];
   BOOL result = FALSE;

   if (snprintf(buffer,
                sizeof(buffer),
                "%s/%s",
                mfsd_config->underlying_fs,
                file) > -1) {
      fd_t fd = creat(buffer, 0666);

      if (fd != -1) {
         if (close(fd) == 0) {
            result = TRUE;
         }
         else {
            perror("close()");
         }
      }
      else {
         perror("creat()");
      }
   }
   else {
      fprintf(stderr, "%s: path too long %s.\n",
              mfsd_config->program_name,
              file);
   }

   return result;
}

static BOOL do_unmigrate(const mfsd_config_t *mfsd_config,
                         const char *path_name,
                         mfs_import_object_locator_t locator,
                         const char *migrated_on) {
   BOOL result = FALSE;
   char buffer[MAX_PATHNAME_LENGTH];

   printf("do_unmigrate: wants file %s located on %ld migrated on %s.\n",
          path_name, locator, migrated_on);

   if (has_extension(migrated_on, ".tar.gz")) {
      /* NOTES
       *    - This is based on the tested tar feature that if the file
       *      exists with a zero file size and chattr A, the new file
       *      has the same inode and the same chattr A bit.
       *    - Locator is unused for this type of migrated data.
       *    - The case: file is migrated and unmigrate already in progress
       *      is not a risk since unmigration is only done by the
       *      single-threaded mfsd.
       * BUGS
       *    - The case: file is BEING currently migrated (by mfs_migrd)
       *      is not really handled (think again, see SPEC BUGS).
       */
      if (snprintf(buffer,
                   sizeof(buffer),
                   "cd %s && tar xvfz %s %s",
                   mfsd_config->underlying_fs,
                   migrated_on, /* supposed absolute name */
                   path_name)) {
         printf("executing %s.\n", buffer);

         if (system(buffer) == 0) {
            result = TRUE;
         }
         else {
            fprintf(stderr, "%s: system(do_unmigrate(1)) failed.\n",
                    mfsd_config->program_name);
         }                   
      }
      else {
	 fprintf(stderr, "%s: snprintf() failed.\n",
                 mfsd_config->program_name);
      }
   }
   else if (!strcmp(migrated_on, MFS_SPECIAL_TEMP_MIGRATED_ON)) {
      /* NOTES
       *    - This is based on the tested cp -a feature that if the
       *      file exists with a zero file size and chattr A, the
       *      new file has the same inode and the same chattr A bit.
       *    - Locator is unused for this type of migrated data.
       *    - The case: file is migrated and unmigrate already in progress
       *      is not a risk since unmigration is only done by the
       *      single-threaded mfsd.
       * BUGS
       *    - The case: file is BEING currently migrated (by mfs_migrd)
       *      is not really handled (think again, see SPEC BUGS).
       */

      if (snprintf(buffer,
                   sizeof(buffer),
                   "cp -a %s/%s %s/%s",
                   mfsd_config->temporary_directory,
                   path_name,
                   mfsd_config->underlying_fs,
                   path_name)) {
         printf("executing %s.\n", buffer);

         if (system(buffer) == 0) {
            remove_file_from_temp(mfsd_config, path_name); /* res. ignored */
            result = TRUE;
         }
         else {
            fprintf(stderr, "%s: system(do_unmigrate(2)) failed.\n",
                    mfsd_config->program_name);
         }                   
      }
      else {
	 fprintf(stderr, "%s: snprintf() failed.\n",
                 mfsd_config->program_name);
      }
   }
   else {
      fprintf(stderr, "%s: unmigration method for %ld/%s unknown.\n",
                 mfsd_config->program_name,
                 locator,
                 migrated_on);
   }

   if (result) {
      /* NOTES
       *    - An optimization could be that we remove from the index
       *      before the first write. At this time we just forget about
       *      anything being unmigrated.
       */
      if (!index_remove(path_name)) {
         fprintf(stderr, "%s: index_remove() failed.\n",
                 mfsd_config->program_name);
      }

      /* The following causes SUBSEQUENT opens to succeed directly. For
       * all the queries in the mfs queues, they are going to be answered
       * with MFS_STATUS_OBJECT_NOT_FOUND, which will cause the underlying
       * fs object to be accessed.
       */
      result = set_unset_migrated(mfsd_config, path_name, FALSE);
   }

   return result;
}

static BOOL has_extension(const char *file_name, const char *extension) {
   BOOL result = FALSE;

   if (strlen(file_name) >= strlen(extension)) {
      if (strcmp(file_name + strlen(file_name) - strlen(extension),
                 extension) == 0) {
         result = TRUE;
      }
   }

   return result;
}

static mfs_index_object_size_t
   traverse_and_migrate(mfsd_config_t *mfsd_config,
                        const char *current_directory_path,
                        mfs_index_object_size_t migrated_bytes) {
   DIR *directory;
   BOOL last_file = FALSE;

   directory = opendir(current_directory_path);
   if (directory) {
      while ((migrated_bytes < MINIMUM_BATCH_THRESHOLD)
             && !last_file) {
         struct dirent *dirent = readdir(directory);

         if (dirent) {
            char file_name[MAX_PATHNAME_LENGTH];
            struct stat stat_buf;

            /* Only look at files for migration and at directories for
             * recursive traversal.
             * ASSUMPTION
             *    - dirent->d_name is NUL terminated.
             */
            if (snprintf(file_name,
                         sizeof(file_name),
                         "%s/%s",
                         current_directory_path,
                         dirent->d_name) > -1) {
               fprintf(stderr, "Examining: %s.\n", file_name);
	       fflush(stderr);

               if (stat(file_name, &stat_buf) == 0) {
                  if (S_ISREG(stat_buf.st_mode)) {
                     if (migration_candidate(mfsd_config,
                                             file_name,
                                             &stat_buf)) {
                        if (migration_do(mfsd_config, file_name)) {
                           migrated_bytes += stat_buf.st_size;
                        } 
                        else {
                           fprintf(stderr,
                                   "%s: %s was a migration candidate, but was "
                                   "not migrated (this may not be an error if "
                                   "another migration process is running, ie "
                                   "the file has been migrated in between).\n",
                                   mfsd_config->program_name,
                                   file_name);
                        }
                     }
                  }
                  else if (S_ISDIR(stat_buf.st_mode)) {
                     if (strcmp("..", dirent->d_name)
                         && strcmp(".", dirent->d_name)) {
                    
			migrated_bytes
			   = traverse_and_migrate(mfsd_config,
						  file_name,
						  migrated_bytes);
                     }
                  }
                  else {
                     printf("File %s of type %d ignored.\n",
                            file_name,
                            stat_buf.st_mode);
                  }
               }
               else {
                  perror("stat()");
                  fprintf(stderr, "%s: stat failed for %s.\n",
                          mfsd_config->program_name,
                          file_name);
               }
            }
            else {
               fprintf(stderr, "%s: file %s/%s, too long file name.\n",
                       mfsd_config->program_name,
                       current_directory_path,
                       dirent->d_name);
            }
         }
         else {
            last_file = TRUE;
         }
      }

      if (closedir(directory)) {
         perror("closedir()");
      }
   }
   else {
      perror("opendir()");
      fprintf(stderr, "%s: the underlying fs %s cannot be listed.\n",
              mfsd_config->program_name,
              current_directory_path);
   }

   return (migrated_bytes);
}

static BOOL do_mass_migration(mfsd_config_t *mfsd_config) {
   return (traverse_and_migrate(mfsd_config,
                                mfsd_config->underlying_fs,
                                0) > 0);
}

static BOOL migration_candidate(mfsd_config_t *mfsd_config,
                                const char *file_name,
                                const struct stat *stat_buf) {
   return (stat_buf->st_size >= MINIMUM_SIZE_MIGRATE);
}

static BOOL migration_do(mfsd_config_t *mfsd_config, const char *file_name) {
   /* We migrate first to temp
    */

   fprintf(stderr, "Migrating: %s.\n", file_name);
   fflush(stderr);

   if (set_unset_migrated(mfsd_config, file_name, TRUE)) {
      if (copy_file_to_temp(mfsd_config, file_name)) {
         mfs_import_data_t import;

 	 if (get_file_size(mfsd_config, file_name, &import.size)) {
	    import.locator = 0;
	    strncpy(import.migrated_on,
                    MFS_SPECIAL_TEMP_MIGRATED_ON,
                    sizeof(import.migrated_on));

	    if (index_add(file_name, &import)) {
	       if (truncate_file(mfsd_config, file_name)) {
		  return TRUE;
	       }

	       index_remove(file_name); /* res. ignored */
	    }
         }
      }

      set_unset_migrated(mfsd_config, file_name, FALSE); /* res. ignored */
      remove_file_from_temp(mfsd_config, file_name); /* res. ignored */
   }

   return FALSE;
}

static BOOL get_file_size(mfsd_config_t *mfsd_config,
                          const char *file_name,
                          mfs_index_object_size_t *size) {
   char buffer[MAX_PATHNAME_LENGTH];

   if (snprintf(buffer,
                      sizeof(buffer),
                      "%s/%s",
                      mfsd_config->underlying_fs,
                      file_name) > -1) {
      struct stat stat_buf;

      if (stat(buffer, &stat_buf) == 0) {
         *size = stat_buf.st_size;
         return TRUE;
      }
      else {
         perror("stat()");
      }
   }
   else {
      fprintf(stderr, "%s: path too long %s.\n",
	      mfsd_config->program_name,
	      file_name);
   }

   return FALSE;
}
                          

static BOOL remove_file_from_temp(const mfsd_config_t *mfsd_config,
                                  const char *file_name) {
   char buffer[MAX_PATHNAME_LENGTH];

   if (snprintf(buffer,
                      sizeof(buffer),
                      "%s/%s",
                      mfsd_config->temporary_directory,
                      file_name) > -1) {
      if (unlink(buffer) == 0) {
         return TRUE;
      }
      else {
         perror("unlink()");
      }
   }
   else {
      fprintf(stderr, "%s: path too long %s.\n",
	      mfsd_config->program_name,
	      file_name);
   }

   return FALSE;
}

static BOOL copy_file_to_temp(mfsd_config_t *mfsd_config,
                              const char *file_name) {
   char *last_slash = strrchr(file_name, '/');
   BOOL result = FALSE;

   if (strlen(file_name) < MAX_PATHNAME_LENGTH) {
      char buffer[MAX_PATHNAME_LENGTH];

      if (last_slash) {
	 if (snprintf(buffer,
                      sizeof(buffer),
		      "%s/%s",
		      mfsd_config->temporary_directory,
		      file_name) > -1) {
            last_slash = strrchr(buffer, '/');
            if (last_slash) {
               *last_slash = '\0';

	       if (create_directory_path(buffer)) {
		  result = TRUE;
	       }
	       else {
		  fprintf(stderr,
			  "%s: can't create directory path %s. Aborting.\n",
			   mfsd_config->program_name,
			   buffer);
	       }
            }
            else {
	       fprintf(stderr, "%s: incoherency.\n",
		       mfsd_config->program_name);
            }
	 }
	 else {
	    fprintf(stderr, "%s: temp path too long %s.\n",
		    mfsd_config->program_name,
		    file_name);
	 }
      }
      else {
         result = TRUE;
      }

      if (result) {
	 if (snprintf(buffer,
                      sizeof(buffer),
		      "cp -a %s/%s %s/%s",
		      mfsd_config->underlying_fs,
		      file_name,
		      mfsd_config->temporary_directory,
		      file_name) > -1) {
            /* ASSUMPTION
             *    - no attr copy.
             */
            result = (system(buffer) == 0);
	 }
	 else {
	    fprintf(stderr, "%s: cp path too long %s.\n",
		    mfsd_config->program_name,
		    file_name);
            result = FALSE;
	 }
      }
   }
   else {
      fprintf(stderr, "%s: path too long %s.\n",
	      mfsd_config->program_name,
	      file_name);
   }

   return result;
}

static BOOL remount_rw_workaround(mfsd_config_t *mfsd_config,
                                  const char *fs,
                                  BOOL please_fork) {
   char buffer[MAX_PATHNAME_LENGTH];
   BOOL result = FALSE;

   if (snprintf(buffer,
                sizeof(buffer),
                "%smount %s -o remount,rw%s",
                please_fork ? "(sleep 5;" : "",
                fs,
                please_fork ? ") &" : "") > -1) {
      if (system(buffer) == 0) {
         result = TRUE;
      }
      else {
         fprintf(stderr, "%s: system(remount_rw_workaround) failed.\n",
                 mfsd_config->program_name);
      }
   }
   else {
      fprintf(stderr, "%s: path too long %s.\n", 
              mfsd_config->program_name,
              fs);
   }

   return result;
}

static BOOL truncate_file(mfsd_config_t *mfsd_config,
                          const char *file_name) {
   int retries = 5;
   char buffer[MAX_PATHNAME_LENGTH];
   BOOL result = FALSE;

   while (!result && retries--) {

      remount_rw_workaround(mfsd_config,
			    mfsd_config->mfs_mount_point,
                            FALSE); /* res. ign. */
      if (!remount_rw_workaround(mfsd_config,
				 mfsd_config->underlying_fs,
                                 FALSE)) {
	 fprintf(stderr, "%s: workaround 1 failed.\n",
                         mfsd_config->program_name);
	 break; /* out of while */
      }
      else if (snprintf(buffer,
		   sizeof(buffer),
		   "%s/%s",
		   mfsd_config->underlying_fs,
		   file_name) > -1) {
	 if (truncate(buffer, 0) == 0) {
	    result = TRUE;
	 }
	 else {
	    perror("truncate()");
	 }
      }
      else {
	 fprintf(stderr, "%s: path too long %s.\n", 
		 mfsd_config->program_name,
		 file_name);
	 break; /* out of while */
      }

      if (!result) {
         sleep(1);
      }
   }

   return result;
}

static BOOL create_directory_path(const char *name) {
   char buffer[MAX_PATHNAME_LENGTH];
   int size = strlen(name);

   if (size >= sizeof(buffer)) {
      fprintf(stderr, "cdp: size too big.\n");
      return FALSE;
   }

   while (size) {
      char *tmp;

      while (size && (name[size - 1] == '/')) {
         size--;
      }

      strncpy(buffer, name, size);
      buffer[size] = '\0';

      if (mkdir(buffer, 0777 /* BUGS
                              *    - This is presumably wrong.
                              */)) {
         if (errno == EEXIST) {
            return TRUE;
         }
         else {
            perror("mkdir()");
            return FALSE;
         }
      }

      tmp = strrchr(buffer, '/');
      if (tmp && (tmp - buffer)) {
         size = (tmp - buffer);
      }
      else {
         size = 0;
      }
   }
   
   return TRUE;
}

static void simplify_path(char *path, const char *mfs_mount_point) {
   BOOL any_action_taken;

   if (!strncmp(path, mfs_mount_point, strlen(mfs_mount_point))
       && path[strlen(mfs_mount_point)] == '/') {
      shift_string(path, strlen(mfs_mount_point) + 1);
   }

   do {
      char *temp;
      any_action_taken = FALSE;

      if (!strncmp(path, "./", 2)) {
         shift_string(path, 2);
         any_action_taken = TRUE;
      }
      else if (!strncmp(path, "../", 3)) {
         /* Ignore this.
          * BUGS
          *    - Should be an error, modify spec.
          */
         shift_string(path, 2);
         any_action_taken = TRUE;
      }
      else if ((temp = string_scan(path, "/./"))) {
         any_action_taken = TRUE;
         shift_string(temp, 2);
      }
      else if ((temp = string_scan(path, "/../"))) {
         char *temp2;
         any_action_taken = TRUE;

         /* This is very special handling */

         /* First remove the expression, leaving only a / at temp */         
         shift_string(temp, 3);

         /* The find the first / starting from before temp */
         if ((temp2 = strchr(temp - 1, '/'))) {
            /* And remove that */
            shift_string(temp2, (temp - temp2));
         }
         else {
            /* BUGS
             *    - Should be an error, modify spec.
             */
            ;
         }
      }
      else if ((temp = string_scan(path, "//"))) {
         any_action_taken = TRUE;
         shift_string(temp, 2);
      }
   } while (any_action_taken);
}

static void shift_string(char *string, int count) {
   if (strlen(string) >= count) {
      while (string[count]) {
	 *string = string[count];
         string++;
      }
      *string = '\0';
   }
}

static char *string_scan(const char *string, const char *part) {
   while (*string) {
      if (!strncmp(string, part, strlen(part))) {
         return (char *) string;
      }
      string++;
   }

   return NULL;
}

static BOOL threads_launch(mfsd_config_t *mfsd_config, mfs_info_t mfs_info) {
   BOOL result = TRUE;

   mfsd_config->children_count = 0;
   if (mfs_info.reserved_entries > mfs_info.queue_depth) {
      fprintf(stderr, "%s: reserved greater than available.\n",
                      mfsd_config->program_name);
      result = FALSE;
   }
   else if (mfs_info.reserved_entries == mfs_info.queue_depth) {
      printf("%s: no children allocated --- synchronous operation.\n",
             mfsd_config->program_name);
   }
   else {
      mfsd_child_number_t children = mfs_info.queue_depth
                                   - mfs_info.reserved_entries;

      mfsd_config->children = calloc(children, sizeof(mfsd_child_t));
      if (mfsd_config->children) {
	 while (children--) {
            if (thread_launch
                   (&(mfsd_config->children[mfsd_config->children_count]))) {
   	       mfsd_config->children_count++;
            }
            else {
	       fprintf(stderr, "%s: thread %d cannot be started.\n",
			       mfsd_config->program_name,
                               mfsd_config->children_count);
               threads_stop(mfsd_config);
	       result = FALSE;
            }
	 }
      }
      else {
	 fprintf(stderr, "%s: no memory for children table.\n",
			 mfsd_config->program_name);
	 result = FALSE;
      }
   }

   return result;
}

static void threads_stop(mfsd_config_t *mfsd_config) {
   mfsd_child_number_t child;

   for (child = 0; child < mfsd_config->children_count; child++) {
      thread_stop(&(mfsd_config->children[child]));
   }

   mfsd_config->children_count = 0;
   free(mfsd_config->children);
   mfsd_config->children = NULL;
}

static BOOL thread_launch(mfsd_child_t *child) {
   int pipes_to[2];
   int pipes_from[2];

   if (pipe(pipes_to) || pipe(pipes_from)) {
      perror("pipe()");
      return FALSE;
   }
   else {
      child->to = pipes_to[1];
      child->from = pipes_from[0];

      fflush(stdout);
      fflush(stderr);

      switch ((child->pid = fork())) {
         case 0: /* child */
            close(pipes_to[1]); /* ign. result */
            close(pipes_from[0]); /* ign. result */
            mfsd_thread(&mfsd_config, pipes_to[0], pipes_from[1]);
            _exit(0);
            break;
        case -1: /* error */
            perror("fork()");
            return FALSE;
            break;
        default: /* father */
           close(pipes_to[0]); /* ign. result */
           close(pipes_from[1]); /* ign. result */

           child->state = MFSD_CHILD_STATE_INITIALIZING;
           child->object_number = MFSD_OBJECT_NUMBER_NONE;

           return thread_initialize(child);
           break;
      }
   }
}

static void thread_stop(mfsd_child_t *child) {
   /* @@@ */
}

static BOOL thread_initialize(mfsd_child_t *child) {
   return FALSE; /* @@@ init protocol, check there, set state @@@ */
}

static BOOL thread_get_request(fd_t in, mfsd_operation_t *operation) {
   return (read(in, operation, sizeof(*operation)) == sizeof(*operation));
}

static BOOL thread_send_result(fd_t out, mfsd_pipe_answer_result_t *result) {
   return (write(out, result, sizeof(*result)) == sizeof(*result));
}

static void mfsd_thread(const mfsd_config_t *mfsd_config,
                        fd_t commands,
                        fd_t answers) {
   mfsd_operation_t operation;
   BOOL had_error = FALSE;

   while (!had_error && thread_get_request(commands, &operation)) {
      mfsd_pipe_answer_result_t result = MFSD_PIPE_ANSWER_RESULT_ERROR;

      if (operation.operation == MFSD_OPERATION_UNMIGRATE) {
         if (do_unmigrate(mfsd_config,
                          operation.path_name,
                          operation.import_data.locator,
                          operation.import_data.migrated_on)) {
            result = MFSD_PIPE_ANSWER_RESULT_GOOD;
         }
      }
      else {
         fprintf(stderr, "thread: unknown command 0x%x.\n",
                         operation.operation);
         fflush(stderr);
      }

      if (!thread_send_result(answers, &result)) {
         had_error = TRUE;
      }
   }

   if (!had_error) {
      /* @@@ proper shutdown protocol @@@ */
   }
}

mfsd_object_t *queue_allocate_object(const mfsd_config_t *mfsd_config) {
   return NULL; /* @@@ */
}

BOOL send_request(const mfsd_config_t *mfsd_config,
                  mfsd_child_t *child,
                  mfsd_object_t *object) {
   return (write(child->to, &object->operation, sizeof(object->operation))
           == sizeof(object->operation));
}

void internal_queue(const mfsd_config_t *mfsd_config,
                    mfsd_object_t *object) {
}

static BOOL queue_request(const mfsd_config_t *mfsd_config,
                          const char *path_name,
                          mfs_import_data_t *import_data) {
   mfsd_child_number_t current_child = 0;
   mfsd_child_t *child = mfsd_config->children;
   mfsd_object_t *object = queue_allocate_object(mfsd_config);

   if (object) {
      object->handler = NULL;
      /* either copy it to waiting queue (this should be done
       * with a warning as should not happen), or to a thread
       */

      while ((current_child < mfsd_config->children_count)
             && (object->handler == NULL)) {
	 if ((child->state != MFSD_CHILD_STATE_NOT_LAUNCHED)
	     && (child->object_number == MFSD_OBJECT_NUMBER_NONE)) {
            object->handler = child;
	 }

	 child++;
	 current_child++;
      }

      if (object->handler) {
         if (!send_request(mfsd_config, object->handler, object)) {
            object->handler = NULL;
         }
      }

      if (object->handler == NULL) {
         internal_queue(mfsd_config, object);
      }

      return TRUE;
   }
   else {
      fprintf(stderr, "can't allocate object.\n");
      return FALSE;
   }

   /* NOT REACHED */
}

static void threads_fd_do(mfsd_config_t *mfsd_config, fd_set *fds) {
   mfsd_child_number_t current_child = 0;
   mfsd_child_t *child = mfsd_config->children;
   BOOL one_freed = FALSE;

   while (current_child < mfsd_config->children_count) {
      if ((child->state != MFSD_CHILD_STATE_NOT_LAUNCHED)
          && FD_ISSET(child->from, fds)) {
         mfsd_pipe_answer_result_t result;

         if (read(child->from, &result, sizeof(result)) == sizeof(result)) {
	    if (child->object_number == MFSD_OBJECT_NUMBER_NONE) {
	       fprintf(stderr, "STRANGE: child %d no object wants to talk.\n",
			       current_child);
	    }
            else {
               fprintf(stderr, "Data available from child %d.\n",
                               current_child);
               /* @@@ */
               one_freed = TRUE; /* @@@ */
            }
         }
         else {
            perror("read())");
         }
      }

      child++;
      current_child++;
   }

   if (one_freed) {
      /* @@@ also process the internal queue since a free slot @@@ */
   }
}

static fd_t threads_fd_set(mfsd_config_t *mfsd_config, fd_set *fds) {
   mfsd_child_number_t current_child = 0;
   fd_t max_fd = 0;
   mfsd_child_t *child = mfsd_config->children;

   while (current_child < mfsd_config->children_count) {
      if ((child->state != MFSD_CHILD_STATE_NOT_LAUNCHED)
          && (child->object_number != MFSD_OBJECT_NUMBER_NONE)) {
         FD_SET(child->from, fds);
         max_fd = MAX(max_fd, child->from);
      }

      child++;
      current_child++;
   }

   return max_fd;
}