#! /bin/sh
* Project mfs --- Migration file system
# test.sh -- Simple regression testing script
# Creation: 29/12/1999
# Update:   29/12/1999
# V1.0 PV001
# AUTHOR
#    Marc SCHAEFER <schaefer@alphanet.ch>
# LICENSE
#    GPL
# DESCRIPTION
#    This implements a very simple test, using locally generated data,
#    from /lib and /etc, of data integrity, unmigration, and
#    import from an already existing tar.gz file.
# NOTES
#    - This does not test re-migration of the data (we do generate a big
#      file in order to re-migrate everything, however we do not
#      unmigrate the data again to verify it's integrity).
# BUGS
# TODO
# BASED-ON
#    Original work.
# MODIFICATION-HISTORY
#    17/10/99  schaefer  Created this file.
# $Id: test.sh,v 1.2 1999/12/30 10:04:16 schaefer Exp $

if [ $# != 5 ]; then
   echo "$0 mfs-mount-point underlying-fs-mount-point temp-directory index-file spool-directory"
   echo "$0: bad args.\n"
   exit 2
fi

MFS_MOUNT_POINT=$1
UNDERLYING_FS=$2
TEMP_DIRECTORY=$3
INDEX_FILE=$4
SPOOL_DIRECTORY=$5

insmod src/mfs/mfs.o
if [ $? != 0 ]; then
   echo "$0: can't insert the mfs filesystem kernel driver module.'
   exit 1
fi

TO_WIPE_DIRECTORIES=$MFS_MOUNT_POINT $TEMP_DIRECTORY $SPOOL_DIRECTORY
TO_WIPE_FILES=$INDEX_FILE

echo "Now wiping the following directories: $TO_WIPE_DIRECTORIES"
echo -n 'Please confirm [y/N] ?:'
read $confirmation
case $confirmation in
   y|Y|[yY][eE][sS]) ;;
   *) echo 'Aborted'
      exit 1;;
esac
rm -rf $TO_WIPE_DIRECTORIES
if [ $? != 0 ]; then
   echo "$0: can't wipe directories $TO_WIPE_DIRECTORIES"
   exit 1
fi

mkdir -p $TO_WIPE_DIRECTORIES
if [ $? != 0 ]; then
   echo "$0: can't make directories $TO_WIPE_DIRECTORIES"
   exit 1
fi

echo "Now wiping the following files: $TO_WIPE_FILES"
echo -n 'Please confirm [y/N] ?:'
read $confirmation
case $confirmation in
   y|Y|[yY][eE][sS]) ;;
   *) echo 'Aborted'
      exit 1;;
esac
rm -f $TO_WIPE_FILES
if [ $? != 0 ]; then
   echo "$0: can't wipe files $TO_WIPE_FILES"
   exit 1
fi

rm -f $UNDERLYING_FS/fake_mfsd && touch $UNDERLYING_FS/fake_mfsd
if [ $? != 0 ]; then
   echo "$0: can't create the special mfsd file"
   exit 1
fi

mount -t mfs -o dir=$UNDERLYING_FS /dev/null $MFS_MOUNT_POINT
if [ $? != 0 ]; then
   echo "$0: can't mount the fs"
   exit 1
fi

./src/mfsd/mfsd $TEMP_DIRECTORY $MFS_MOUNT_POINT $UNDERLYING_FS $INDEX_FILE &
if [ $? != 0 ]; then
   echo "$0: can't mount the fs"
   exit 1
fi

echo ""
echo "mfs fs is now active (mfs is under $MFS_MOUNT_POINT)"
echo ""

DIRECTORIES=/lib /etc
COUNT=0
for i in $DIRECTORIES
do
   echo "Now creating a tar.gz file in $SPOOL_DIRECTORY with some random files"
   echo "from $i. This simulates files migrated on some off-line storage."

   (cd $i && find . \( -type f -o -type d \) -print | tar -cfz - -T - > $SPOOL_DIRECTORY/spool-tar-file-$COUNT.tar.gz)
   if [ $? != 0 ]; then
      echo "$0: can't create the tar file in $i."
      exit 1
   fi

   COUNT=`expr $COUNT + 1`
   echo ""
done

COUNT=0
for i in $DIRECTORIES
do
   echo "Now importing the migrated file inode information into the mfs"
   echo "filesystem, as migrated data, with the mfs_import(8) command."
   echo "We do that now for data originally from $i."

   tar -tvfz $SPOOL_DIRECTORY/spool-tar-file-$COUNT.tar.gz | ./src/mfs_import/mfs_import $MFS_MOUNT_POINT $SPOOL_DIRECTORY/spool-tar-file-$COUNT.tar.gz
   if [ $? != 0 ]; then
      echo "$0: can't import the tar file originally from $i."
      exit 1
   fi

   COUNT=`expr $COUNT + 1`
   echo ""
done

COUNT=0
for i in $DIRECTORIES
do
   echo "Now comparing the spooled archive with the mfs filesystem."
   echo "This will cause ALL migrated files originally coming from"
   echo "the archive $i to be unmigrated, and possibly at the"
   echo "end the spooled archive to be deleted."

   (cd $MFS_MOUNT_POINT && tar --compare -zf $SPOOL_DIRECTORY/spool-tar-file-$COUNT.tar.gz)
   if [ $? != 0 ]; then
      echo "$0: can't import the tar file originally from $i."
      exit 1
   fi

   if [ -f $SPOOL_DIRECTORY/spool-tar-file-$COUNT.tar.gz ]; then
      echo "$0: the spool file for $i was not deleted."
      echo "$0: [Not yet implemented, ignore this warning]."
      # Think again. Maybe mfs deletes it only if all the contained files
      # are modified ?  As of today it's not implemented. Anyway, for
      # tapes, reclaimation will do that.
      # exit 1
   fi

   COUNT=`expr $COUNT + 1`
   echo ""
done

echo "So, you survived upto now. The situation is as follows: there are"
echo "plenty of (unmigrated) files in $MFS_MOUNT_POINT. We are now creating"
echo "a dummy file so that space gets low and that some data is migrated."
echo "As of current implementation, we can only have the max file size"
echo "upto around the max fs size, since we can only migrate NON active"
echo "complete files."
echo ""

MAX_FS_SIZE=`df $MFS_MOUNT_POINT | tail -1 | awk '{ print $2 }'`

echo -n "Creating a dummy file of `echo $MAX_FS_SIZE / 1024 | bc` MBs ... "
# BUGS
#    - Maybe too big.
dd if=/dev/zero of=$MFS_MOUNT_POINT/dummy-file bs=1024 count=$MAX_FS_SIZE
if [ $? != 0 ]; then
   echo "error"
   echo "$0: can't create the big dummy file.\n"
   exit 1
fi

echo "done"
echo ""

echo "Ok, very nice. In theory now you should be able to re-compare all the"
echo "tar files again (if they still exist ...), and that should migrate"
echo "our nice big files. We will try that another day."

exit 0